#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import tarfile

from BaseSystem import BaseSystem
from stratuslab.system.PackageInfo import PackageInfo
import stratuslab.Util as Util
from stratuslab import Exceptions
import re

packageManagerReposConfDir = '/etc/yum.repos.d'
installCmd = 'yum -y --nogpgcheck install'
updateCmd = 'yum update'
cleanPackageCacheCmd = 'yum clean all'
queryPackageCmd = 'rpm -q'

repoFileNamePattern = '%s/%s.repo' %(packageManagerReposConfDir, '%s')

class CentOS(BaseSystem):

    os = 'centos'

    def __init__(self):
        super(CentOS, self).__init__()
        
        self.repoFileNamePattern = repoFileNamePattern
        
        self.systemName = 'CentOS 5.5'
        self.arch = self.getSystemArch()
        self.installCmd = installCmd
        self.remotePackages = [
            # ('name', 'version', 'arch', 'uri')
            # --> <uri>/<name>-<version>.<arch>.rpm
            ('scons', '1.2.0-1', 'noarch', 'http://prdownloads.sourceforge.net/scons'),
            ('xmlrpc-c', '1.06.18-1.el5.kb', self.arch, 'http://centos.karan.org/el5/extras/testing/%(arch)s/RPMS'),
            ('xmlrpc-c-devel', '1.06.18-1.el5.kb', self.arch, 'http://centos.karan.org/el5/extras/testing/%(arch)s/RPMS'),
        ]
#        self.remoteSources = [
#            # ('name', 'version', 'uri', 'extension')
#            # --> <uri>/<name>-<version>.<extention>
#            ('git', '1.7.1.1', 'http://kernel.org/pub/software/scm/git', 'tar.gz'),
#            ('sqlite-amalgamation', '3.6.17', 'http://www.sqlite.org', 'tar.gz'),
#        ]
        self.frontendDeps = [
            'openssh', 'ruby', 'gcc', 'gcc-c++', 'zlib-devel', 'genisoimage', 'curl'
        ]
        self.nodeDeps = ['ruby', 'curl', 'libvirt', 'genisoimage', 'openssh', 'bridge-utils',
                         'stratuslab-libvirt-monitor']
        self.hypervisorDeps = {
            'xen': ['xen', 'kernel-xen'],
            'kvm': ['qemu-kvm'],
        }
        self.fileSharingFrontendDeps = {
            'nfs': [],
            'ssh': [],
        }
        self.fileSharingNodeDeps = {
            'nfs': [],
            'ssh': [],
        }

        self.extraRepos = {
            'epel' : {'content' : """[epel]
name=Extra Packages for Enterprise Linux 6 - $basearch
#baseurl=http://download.fedoraproject.org/pub/epel/6/$basearch
mirrorlist=https://mirrors.fedoraproject.org/metalink?repo=epel-6&arch=$basearch
failovermethod=priority
enabled=1
gpgcheck=0
#gpgkey=file:///etc/pki/rpm-gpg/RPM-GPG-KEY-EPEL-6
""", 'filename' : repoFileNamePattern % 'epel'},

            self.caRepoName : {'content' : """[EGI-trustanchors]
name=EGI-trustanchors
baseurl=http://repository.egi.eu/sw/production/cas/1/current/
#gpgkey=http://repository.egi.eu/sw/production/cas/1/GPG-KEY-EUGridPMA-RPM-3
gpgcheck=0
enabled=1
""", 'filename' : repoFileNamePattern % self.caRepoName}
            }

        self.packages = {'apache2': PackageInfo('httpd','/etc/httpd'),
                         'dhcp': PackageInfo('dhcp',
                                             configFile='/etc/dhcp/dhcpd.conf',
                                             initdScriptName='dhcpd'),
                         'CA' : PackageInfo('ca-policy-egi-core', 
                                                repository=self.caRepoName),
                         'MySQLServer': PackageInfo('mysql-server',
                                                    initdScriptName='mysqld'),
                         'fetch-crl': PackageInfo('fetch-crl',
                                                  packageVersion='3*'),
                         'sendmail': PackageInfo('sendmail')}

        self.installPackagesErrorMsgs = ['No package .* available']


    def getSystemArch(self):
        try:
            _, _, _, _, arch = os.uname()
            if arch == 'x86_64':
                return arch
            else:
                return 'i386'
        except AttributeError:
            return 'i386'

    # -------------------------------------------
    #     Package manager and related
    # -------------------------------------------
    
    def addRepositories(self, packages):
        """Accepts package names and aliases as defined in self.packages.
        """
        repos = []
        for pkgName in packages:
            repo = ''
            if pkgName in self.packages:
                repo = self.packages[pkgName].repository
            else:
                for pkgInfo in self.packages.values():
                    if pkgInfo.packageName == pkgName:
                        repo = pkgInfo.repository
            if repo and repo not in repos:
                repos.append(repo)
        
        for repo in repos:
            if repo in self.extraRepos:
                filename = self.extraRepos[repo]['filename']
                content = self.extraRepos[repo]['content']
                Util.filePutContent(filename, content)
            else:
                Util.printError("Repository '%s' is not defined in the extra list of repositories (%s)." % \
                                (repo, ', '.join(self.extraRepos)))

    def updatePackageManager(self):
        pass

    def installFrontendDependencies(self):
        super(CentOS, self).installFrontendDependencies()
        self.installRemotePackages(self.remotePackagesAddress(self.remotePackages))
        self.installSourceDependencies(self.remoteSourcesAddress(self.remoteSources))

    def remotePackagesAddress(self, packages):
        remotePackages = []

        for name, version, arch, uri in packages:
            remotePackages.append('%s/%s-%s.%s.rpm' % (uri % ({'arch': arch}), name, version, arch))

        return remotePackages

    def installRemotePackages(self, packages):
        rpmName = '/tmp/stratus-%d.rpm'
        pkgList = []

        for i in range(len(packages)):
            pkg = packages.pop()
            Util.wget(pkg, rpmName % i)
            pkgList.append(rpmName % i)

        self.installPackages(pkgList)

    def getIsPackageInstalledCommand(self, package):
        cmd = '%s %s' % (queryPackageCmd, package)
        return cmd

    def enableServiceOnBoot(self, service, level='3'):
        cmd = ['chkconfig', '--level', str(level), service, 'on']
        rc, output = self._executeWithOutput(cmd)
        if rc != 0:
            Util.printDetail(output)
        return rc

    # -------------------------------------------
    #     Source build and installation methods
    # -------------------------------------------

    def remoteSourcesAddress(self, sources):
        remoteSources = []

        for name, version, uri, ext in sources:
            remoteSources.append('%s/%s-%s.%s' % (uri, name, version, ext))

        return remoteSources

    def installSourceDependencies(self, sourcesAddress):
        for dep in sourcesAddress:
            self.buildAndInstall(dep)

    def buildAndInstall(self, sourceAddr):
        archive = '/tmp/stratus-deps-src.tar.gz'
        Util.wget(sourceAddr, archive)
        tar = tarfile.open(archive)
        srcFile = tar.getmembers()

        for elem in srcFile:
            tar.extract(elem)

        os.chdir(srcFile[0].name)
        self._execute(['./configure'])
        self._execute(['make', '-j2', 'install'])
        os.chdir('../')

    # -------------------------------------------
    #     File sharing related methods
    # -------------------------------------------

    def configureNewNfsServer(self, mountPoint, networkAddr, networkMask):
        super(CentOS, self).configureNewNfsServer(mountPoint, networkAddr, networkMask)
        self.executeCmd(['service', 'nfs', 'start'])

    # -------------------------------------------
    #     Hypervisor related methods
    # -------------------------------------------

    def _configureKvm(self):
        super(CentOS, self)._configureKvm()
        self.executeCmd(['/etc/init.d/libvirtd start'])
        # Sleep to give a chance to libvirt to create the libvirt-sock
        Util.sleep(5)
        self.executeCmd(['usermod', '-G', 'kvm', '-a', self.oneUsername])
        self.executeCmd(['chown', 'root:kvm',
                        '/var/run/libvirt/libvirt-sock'])
        self.executeCmd(['chmod', 'g+r+w', '/var/run/libvirt/libvirt-sock'])

        self.executeCmd('ln -s /usr/libexec/qemu-kvm /usr/bin/qemu-kvm'.split())

    # -------------------------------------------
    # Network related methods
    # -------------------------------------------

    def _configureNetworkInterface(self, device, ip, netmask):
        deviceConf = '/etc/sysconfig/network-scripts/ifcfg-%s' % device
        data = """DEVICE=%s
IPADDR=%s
NETMASK=%s
""" % (device, ip, netmask)
        Util.filePutContent(deviceConf, data)

    def _persistRemoteBridgeConfig(self, iface, bridge):
        netScriptsDir = '/etc/sysconfig/network-scripts/'
        ifaceFile = '%s/ifcfg-%s' % (netScriptsDir, iface)
        bridgeFile = '%s/ifcfg-%s' % (netScriptsDir, bridge)

        if self._nodeShell('[ -f %s ]' % bridgeFile, shell=True) == 0:
            Util.printWarning(('Bridge configuration already present (%s:%s).' % 
                               (self.nodeAddr, bridgeFile)) + \
                               ' Please update it manually if required.')
            return

        cmd = 'cat %s' % ifaceFile
        rc, output = self._nodeShell(cmd, withOutput=True, shell=True)
        if rc != 0:
            Util.printWarning('Failed to get content of %s:%s:\n%s' % 
                              (self.nodeAddr, ifaceFile, output))
            return

        ifaceConfOrig = output
        try:
            bridgeConf, ifaceConf = self._buildBridgeAndIfaceConfig(ifaceConfOrig,
                                                                    iface, bridge)
        except Exceptions.ConfigurationException, ex:
            Util.printWarning('Failed to build config for %s and %s.\n%s' % 
                              (iface, bridge, str(ex)))
            return
        else:
            self._writeToFilesRemote([(bridgeFile, bridgeConf), (ifaceFile, ifaceConf)])

    @staticmethod
    def _buildBridgeAndIfaceConfig(ifaceConfOrig, iface, bridge):
        
        res = re.search('(^BOOTPROTO.*)', ifaceConfOrig, re.M)
        if not res:
            raise Exceptions.ConfigurationException('BOOTPROTO not defined in %s config.' % iface)
        else:
            BOOTPROTO = res.group()

        bridgeConf = """DEVICE=%s
TYPE=Bridge
ONBOOT=yes
DELAY=0
IPV6INIT=yes
""" % bridge
        
        BOOTPROTO = BOOTPROTO.split('=')[1]
        if BOOTPROTO == 'static':
            # copy static conf from main interface
            bridgeConf += "BOOTPROTO=static\n"
            pat = '.*|'.join(['BROADCAST', 'IPADDR', 'NETMASK', 
                              'NETWORK', 'GATEWAY']) + '.*'
            matches = re.findall('^(%s)' % pat, ifaceConfOrig, re.M)
            for match in matches:
                bridgeConf += '%s\n' % match
        else:
            bridgeConf += "BOOTPROTO=%s\n" % BOOTPROTO

        ifaceConf = """DEVICE=%s
ONBOOT=yes
BRIDGE=%s
""" % (iface, bridge)

        for key in ['IPV6INIT', 'HWADDR']:
            res = re.search('(^%s.*)' % key, ifaceConfOrig, re.M)
            if res:
                ifaceConf += '%s\n' % res.group()
        
        return bridgeConf, ifaceConf


    # -------------------------------------------
    # Security
    # -------------------------------------------
        
    def _enableFetchCrl(self):

        Util.printDetail('Enabling fetch-crl-cron.')
        self.startService('fetch-crl-cron')
        self.enableServiceOnBoot('fetch-crl-cron', '3')


system = CentOS()



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2012, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import commands

from stratuslab.Util import printWarning
from stratuslab.messaging.MsgBase import MsgBase
from stratuslab.HttpClient import HttpClient

class _RestPublisherHttpClient(MsgBase):
    def __init__(self, configHolder):
        self.restEndpoints = '{}'
        super(_RestPublisherHttpClient, self).__init__(configHolder)

        self.resource = '%s/%s' % (self.msg_endpoint, self.msg_queue)
        self.httpClient = HttpClient(configHolder)
        self._assignCredentials()

    def _assignCredentials(self):
        endpts_creds = json.loads(self.restEndpoints)
        try:
            creds = endpts_creds[self.msg_endpoint]
        except KeyError:
            printWarning('WARNING: no matching credentials for ' + self.msg_endpoint)
        else:
            self.httpClient.addCredentials(creds['username'], creds['password'])

    def send(self, message):
        self.httpClient.put(self.resource, message)

class _RestPublisherCurl(MsgBase):
    def __init__(self, configHolder):
        self.restEndpoints = '{}'
        super(_RestPublisherCurl, self).__init__(configHolder)

        self.resource = '%s/%s' % (self.msg_endpoint, self.msg_queue)
        self._assignCredentials()

    def _assignCredentials(self):
        endpts_creds = json.loads(self.restEndpoints)
        try:
            creds = endpts_creds[self.msg_endpoint]
        except KeyError:
            printWarning('WARNING: no matching credentials for ' + self.msg_endpoint)
        else:
            self.username = creds['username']
            self.password = creds['password']

    def send(self, message):
        user = ''
        if self.username and self.password:
            user = '--user %s:%s' % (self.username, self.password)
        cmd = 'curl -d %s -X PUT %s %s' % (message, user, self.resource)
        rc, output = commands.getstatusoutput(cmd)
        if rc != 0 or ('Error' in output):
            raise Exception("Command : %s\nFailed with :\n%s" % (cmd, output))

# FIXME: RestPublisherHttpClient.send() fails at the moment
RestPublisher = _RestPublisherCurl



#
# Copyright (c) 2012, Centre National de la Recherche Scientifique (CNRS)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64
import json
import os

from gzip import GzipFile
from StringIO import StringIO

from contextlib import closing

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

SIZE_ERROR_MSG = 'multipart size (%d) exceeds limit (%d)'

'''
The maximum number of bytes that can be passed as user data is
limited (essentially by AWS).  Having such a limit avoids DOS
attacks; enforce the limit here. 
'''
MAX_BYTES = 16384

'''
This creates a single multipart file from the given parts.  Each part
must be a tuple containing mime-type and the raw contents.  The 
mime-type should be the subtype ONLY.  All files are assumed to be 
text files.
'''
def createMultipartString(parts):
    msg = MIMEMultipart()
    index = 0
    for mimetype, content in parts:
        part = createTextPart(content, mimetype, ('part-%s' % index))
        index = index + 1
        msg.attach(part)
        if (mimetype == 'none'):
            return content
        
    return msg.as_string()

'''
This creates a single multipart file from the given parts.  Each part
must be a tuple containing mime-type and the file to include.  The
mime-type should be the subtype ONLY.  All files are assumed to be
text files.
'''
def createMultipartStringFromFiles(parts):
    msg = MIMEMultipart()
    for mimetype, file in parts:
        with open(file, 'rb') as f:
            contents = f.read()
            part = createTextPart(contents, mimetype, os.path.basename(file))
            msg.attach(part)
            if (mimetype == 'none'):
                return contents

    return msg.as_string()


'''
Create a text message part with the given content, mimetype, and
filename.
'''
def createTextPart(content, mimetype, filename):
    part = MIMEText(content, mimetype)
    part.add_header('Content-Disposition', 'attachment', filename=filename)
    return part


'''
This gzips the multipart content and then base64 encodes the result.
The result is compatible with sending as a value in the context 
key-value pairs supported by OpenNebula.  Note that this checks that
the maximum size of the result does not exceed 16384 bytes.
'''
def encodeMultipart(multipart):
    with closing(StringIO()) as buffer:
        with closing(GzipFile('', 'wb', 9, buffer)) as f:
            f.write(multipart)
        gzipped_data = buffer.getvalue()

    gzipped_size = len(gzipped_data)
    if (gzipped_size > MAX_BYTES):
        raise ValueError(SIZE_ERROR_MSG % (gzipped_size, MAX_BYTES))

    return base64.b64encode(gzipped_data)

'''
Decodes the multipart content and returns it.  The input must be 
a base64 encoded, gzipped file.  It will only decode values that 
do not exceed the maximum size.
'''
def decodeMultipart(encoded_multipart):
    gzipped_data = base64.b64decode(encoded_multipart)
    gzipped_size = len(gzipped_data)
    if (gzipped_size > MAX_BYTES):
        raise ValueError(SIZE_ERROR_MSG % (gzipped_size, MAX_BYTES))

    with closing(StringIO(gzipped_data)) as buffer:
        with closing(GzipFile('', 'rb', 9, buffer)) as f:
            return f.read()

'''
Decodes the multipart content and returns it as JSON.
'''
def decodeMultipartAsJson(dsmode, encoded_multipart):
    info = {}
    if encoded_multipart :
        info['user-data'] = decodeMultipart(encoded_multipart)
    info['dsmode'] = dsmode

    return json.dumps(info)

'''
Creates an authorized_keys file from the given key file(s).  The
result has one key per line (assuming that the input in on one
line.
'''
def createAuthorizedKeysFromFiles(keyfiles):
    with closing(StringIO()) as buffer:
        for keyfile in keyfiles:
            with open(keyfile, 'rb') as f:
                key = f.read()
                buffer.write(key.strip())
            buffer.write("\n")
        return buffer.getvalue()


'''
Creates the encoded contents for the user data given a list of
arguments.  Each argument must be a mimetype and filename pair
separated by a comma.  This treats all of the mimetypes except the
pseudo-mimetype of 'ssh'.
'''
def encodedUserData(args):
    mimefiles = []
    for entry in args:
        mimetype, file = entry.split(',')
        if (mimetype != 'ssh'):
            mimefiles.append((mimetype, file))
    if (len(mimefiles) == 0):
        return None
    else:
        return encodeMultipart(createMultipartStringFromFiles(mimefiles))


'''
Create the encoded contents for the authorized key file from the given
list of arguments.  Each argument must be a mimetype and filename pair
separated by a comma.  This ignores all entries except those with the
pseudo-mimetype of 'ssh'.
'''
def encodedAuthorizedKeysFile(args):
    keyfiles = []
    for entry in args:
        mimetype, file = entry.split(',')
        if (mimetype == 'ssh'):
            keyfiles.append(file)
    if (len(keyfiles) == 0):
        return None
    else:
        contents = createAuthorizedKeysFromFiles(keyfiles)
        return base64.b64encode(contents)

'''
Creates the contents of the context file with the given list of
arguments.  The method accepts an optional separator character, which
is a newline by default. Each argument must be a mimetype and filename
pair separated by a comma.
'''
def contextFile(args, separator="\n"):
   authorized_keys = encodedAuthorizedKeysFile(args)
   user_data = encodedUserData(args)

   kvpairs = []

   # Do NOT add spaces around the equals sign.  These will leak into the
   # values defined in the context file defined by OpenNebula.
   kvpairs.append('CONTEXT_METHOD=cloud-init')
   if authorized_keys:
       kvpairs.append("CLOUD_INIT_AUTHORIZED_KEYS=%s" % authorized_keys)
   if user_data:
       kvpairs.append("CLOUD_INIT_USER_DATA=%s\n" % user_data)

   return separator.join(kvpairs)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from Util import assignAttributes

class CloudInfo(object):
    
    def __init__(self):
        self.attribs = {}
        
    def populate(self, element):
        self._populate(element)
        assignAttributes(self, self.attribs)

    def _populate(self, element, parentHierachy=[]):
        children = self._getChildren(element)
        if children:
            _parentHierachy = self._updateHierachy(element, parentHierachy)
            for child in children:
                self._populate(child, _parentHierachy)
        else:
            # skip the root element
            hierachy = parentHierachy[1:] + [element.tag]
            attributeName = '_'.join(hierachy)
            if isinstance(element.text, unicode):
                text = element.text.encode('utf-8')
            else:
                text = element.text
            self.attribs.__setitem__(attributeName.lower(), text)
        return

    def _getChildren(self, parent):
        return parent.getchildren()
    
    def _updateHierachy(self, element, parentHierachy):
        _parentHierachy = parentHierachy[:]
        _parentHierachy.append(element.tag)
        return _parentHierachy

    def getAttributes(self):
        return self.attribs

    def set(self, key, value):
        setattr(self, key, value)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import re
import string
import time
import copy

import stratuslab.Util as Util
from stratuslab.ConfigHolder import ConfigHolder
from stratuslab.Exceptions import ExecutionException

etree = Util.importETree()

NS_RDF     = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
NS_DCTERMS = 'http://purl.org/dc/terms/'
NS_SLTERMS = 'http://mp.stratuslab.eu/slterms#'
NS_SLREQ   = 'http://mp.stratuslab.eu/slreq#'

imageTypes = ['base', 'grid']
imageKinds = ('machine', 'disk')
imageFormats = ['raw', 'qcow2']

class ManifestInfo(object):

    NS_RDF     = NS_RDF
    NS_DCTERMS = NS_DCTERMS
    NS_SLTERMS = NS_SLTERMS
    NS_SLREQ   = NS_SLREQ

    IMAGE_VALIDITY = (356/2) * 24 * 3600 # 1/2 year in sec

    MANDATORY_CHECKSUMS = ('sha1',)
    CHECKSUM_NAMES = ('md5','sha1','sha256','sha512')

    DISKS_BUS_DEFAULT = 'ide'
    INBOUND_PORTS_DEFAULT = []

    def __init__(self, configHolder=ConfigHolder()):

        self.os = ''
        self.osversion = ''
        self.arch = ''
        self.type = '' # image type: base, grid, ..
        self.version = '' # image version

        self.created = '' # image creation time (in iso8601)
        self.user = self.creator = '' # full name of image creator

        self.email = '' # email address of endorser

        self.compression = '' # image compression: gz, bz2, ..
        self.title = ''
        self.tag = ''
        self.comment = ''
        self.filename = '' # filename of compressed image (old manifest)

        self.deprecated = ''

        self.locations = [] # list of image URIs
        self._locations_xml = '<slterms:location>%(location)s</slterms:location>'

        self.kind = '' # image kind: machine, disk

        self.format = '' # image format: raw, qcow2, ..

        self.bytes = '0' # size of the uncompressed image on disk (in bytes)
        self.md5 = ''
        self.sha1 = ''
        self.sha256 = ''
        self.sha512 = ''

        self.valid = '' # ManifestInfo.created + ManifestInfo.IMAGE_VALIDITY (in iso8601)

        self.identifier = '' # base64 of int(sha1_hex, 16)
        self.serialnumber = ''
        self.hypervisor = '' # kvm, xen, ..

        self.publisher = 'StratusLab'

        self.disksbus = ManifestInfo.DISKS_BUS_DEFAULT
        self.inboundports = ManifestInfo.INBOUND_PORTS_DEFAULT

        self.verboseLevel = 0

        configHolder.assign(self)

        self._template = os.path.join(os.path.join(Util.getTemplateDir(),
                                                   'manifest.xml.tpl'))

        self._manifestTemplateElements = \
                (('type',       'type',       NS_DCTERMS, None),
                 ('created',    'created',    NS_DCTERMS, None),
                 ('valid',      'valid',      NS_DCTERMS, None),
                 ('title',      'title',      NS_DCTERMS, ''),
                 ('tag',        'alternative',NS_DCTERMS, ''),
                 ('comment',    'description',NS_DCTERMS, None),
                 ('compression','compression',NS_DCTERMS, None),
                 ('creator',    'creator',    NS_DCTERMS, self.creator),
                 ('user',       'creator',    NS_DCTERMS, self.creator),
                 ('format',     'format',     NS_DCTERMS, self.format),
                 ('publisher',  'publisher',  NS_DCTERMS, self.publisher),
                 ('os',         'os',         NS_SLTERMS, None),
                 ('arch',       'os-arch',    NS_SLTERMS, None),
                 ('osversion',  'os-version', NS_SLTERMS, None),
                 ('version',    'version',    NS_SLTERMS, None),
                 ('kind',       'kind',       NS_SLTERMS, self.kind),
                 ('disksbus',   'disks-bus',  NS_SLTERMS, self.disksbus),
                 ('hypervisor', 'hypervisor', NS_SLTERMS, self.hypervisor),
                 ('email',      'email',      NS_SLREQ,   self.email))

    @staticmethod
    def addElementToManifest(root, elementName, value):
        elems = root.findall('.//{%s}%s' % (ManifestInfo.NS_SLTERMS, elementName))

        if elems:
            # Check if already present. If yes, do nothing.
            for elem in elems:
                if elem.text and elem.text == value:
                    Util.printWarning("Element '%s' with value '%s' already defined in the manifest. Skipping addition of the element." % \
                                      (elementName, elem.text))
                    return

        elem = etree.Element('{%s}%s' % (ManifestInfo.NS_SLTERMS, elementName))
        elem.text = value

        descriptionElement = root.find('.//{%s}Description' % ManifestInfo.NS_RDF)
        descriptionElement.append(elem)

    @staticmethod
    def addElementToManifestFile(manifestfile, elementName, value):
        """TODO: check if given element is a valid one according to schema."""
        root = etree.ElementTree(file=manifestfile)

        ManifestInfo.addElementToManifest(root, elementName, value)

        root.write(manifestfile)

    def parseManifest(self, manifest):

        try:
            xml_tree = etree.fromstring(manifest)
        except SyntaxError, ex:
            raise ExecutionException('Unable to parse manifest: %s\nMANIFEST:\n%s' %
                                     (str(ex), manifest))

        self.parseManifestFromXmlTree(xml_tree)

    def _setEndorserFromXmlTree(self, xml_tree):
        xpathPrefix = './/{http://mp.stratuslab.eu/slreq#}%s/{http://mp.stratuslab.eu/slreq#}'
        self.endorser = xml_tree.findtext(xpathPrefix % 'endorser' + 'email')

    def _setRequiredAttributesFromXmlTree(self, xml_tree):
        """Attributes required by Schema."""

        for elem, ns in [('identifier', NS_DCTERMS), ('bytes', NS_SLREQ)]:
            try:
                val = getattr(xml_tree.find('.//{%s}%s' % (ns, elem)), 'text')
            except AttributeError:
                raise ExecutionException(
                    "Missing mandatory element '%s' in namespace '%s'" % (elem, ns))
            else:
                attr = elem
                setattr(self, attr, val)

        checksums = xml_tree.findall('.//{%s}checksum' % NS_SLREQ)
        for checksum in checksums:
            checkSumType = checksum.find('.//{%s}algorithm' % NS_SLREQ).text
            checkSumType = _normalizeForInstanceAttribute(checkSumType)
            checkSumValue = checksum.find('.//{%s}value' % NS_SLREQ).text
            setattr(self, checkSumType, checkSumValue)


    def _setClientXmlTemplateAttributesFromXmlTree(self, xml_tree):
        """Attributes from client XML template."""

        for attrObj, elemXml, ns, default in self._manifestTemplateElements:
            try:
                attrVal = getattr(xml_tree.find('.//{%s}%s' % (ns, elemXml)), 'text')
            except AttributeError:
                if default != None:
                    attrVal = default
                else:
                    raise ExecutionException("Missing element '%s' in namespace '%s'" % (elemXml, ns))
            else:
                setattr(self, attrObj, attrVal)


    def _setLocationFromXmlTree(self, xml_tree):
        locations = xml_tree.findall('.//{%s}location' % NS_SLTERMS)
        for location in locations:
            uri = getattr(location, 'text', self.locations)
            if uri and uri not in self.locations:
                self.locations.append(uri)

    def _setInboundPortsFromXmlTree(self, xml_tree):
        ports = xml_tree.findall('.//{%s}inbound-port' % NS_SLTERMS)
        for port in ports:
            if hasattr(port, 'text') and port.text not in self.inboundports:
                self.inboundports.append(port.text)

    def parseManifestFromXmlTree(self, xml_tree):

        self._setEndorserFromXmlTree(xml_tree)
        self._setRequiredAttributesFromXmlTree(xml_tree)
        self._setClientXmlTemplateAttributesFromXmlTree(xml_tree)
        self._setLocationFromXmlTree(xml_tree)
        self._setInboundPortsFromXmlTree(xml_tree)

        self.created = getattr(xml_tree.find('.//{%s}endorsement/{%s}created' % (NS_SLREQ, NS_DCTERMS)), 'text',
                            self.created)

    def parseManifestFromFile(self, filename):
        manifest = file(filename).read()
        self.parseManifest(manifest)

    def buildAndSave(self, filename=''):
        manifestText = self.build()
        if not filename:
            filename = '%s-%s-%s-%s-%s%s' % (self.os, self.osversion,
                                             self.arch, self.type,
                                             self.version, Util.manifestExt)
        Util.filePutContent(filename, manifestText)
        Util.printDetail("Manifest: %s"%filename, self.verboseLevel,
                         Util.VERBOSE_LEVEL_DETAILED)

    def build(self):
        self.created = Util.getTimeInIso8601()
        self.valid = Util.toTimeInIso8601(time.time() + self.IMAGE_VALIDITY)
        identifier = ManifestIdentifier()
        self.identifier = identifier.sha1ToIdentifier(self.sha1)
        return self.tostring()

    def tostring(self):
        template = open(self._template).read()
        self._updateLocationsXml()
        return template % self.__dict__

    def _updateLocationsXml(self):
        self._locations_xml = self._getLocationsAsXml()

    def _getLocationsAsXml(self):
        joint = '\n        '
        return joint.join([copy.copy(self._locations_xml) % {'location':location}
                           for location in self.locations])

    def __str__(self):
        sortedKeys = []
        for k in self.__dict__:
            if not k.startswith('_') and not callable(k):
                sortedKeys.append(k)
        sortedKeys.sort()

        output = '* %s:\n' % self.__class__.__name__
        for k in sortedKeys:
            output += '  %s = %s\n' % (k, self.__dict__[k])
        return output

class ManifestIdentifier(object):
    encoding = string.ascii_uppercase + string.ascii_lowercase + string.digits + '-_'
    encoding = [x for x in encoding]

    decoding = {}
    for i,v in enumerate(encoding):
        decoding[v] = long(i)

    sha1Bits = 160
    fieldBits = 6
    identifierChars = sha1Bits / fieldBits + 1
    divisor = long(2) ** fieldBits

    def sha1ToIdentifier(self, sha1):
        try:
            sha1 = long(sha1, 16)
            sb = ''
            for _ in range(self.identifierChars):
                values = divmod(sha1, self.divisor)
                sha1 = values[0]
                sb += self.encoding[int(values[1])]
            return sb[::-1]
        except ValueError as e:
            raise ValueError('invalid SHA-1 checksum: %s' % sha1)

    def identifierToSha1(self, identifier):
        sha1 = long(0)
        for i in range(len(identifier)):
            sha1 = sha1 << self.fieldBits
            bits = long(self.decoding[identifier[i:i+1]])
            sha1 = sha1 | bits
        return hex(sha1).lstrip('0x').rstrip('L')

def _normalizeForInstanceAttribute(attr):
    return _removeDashes(attr.lower())

def _removeDashes(string):
    return re.sub('-', '', string)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import stratuslab.Util as Util
from stratuslab.ConfigHolder import ConfigHolder
from stratuslab import Exceptions

class Configurable(object):
    def __init__(self, configHolder):
        configHolder.assign(self)
        self._verify()

    @staticmethod
    def convertToSectionDict(config):
        dicts = {}
        for section in config.sections():
            _dict = {}
            for k,v in config.items(section):
                _dict[k] = v
            dicts[section] = _dict
        return dicts

    def parseConfig(self, configFileName):
        return ConfigHolder.parseConfig(configFileName)

    def _verify(self):
        attrs = ['verboseLevel']
        for attr in attrs:
            if not hasattr(self, attr):
                raise Exceptions.ValidationException("%s: missing attribute '%s'" % \
                                                     (self.__class__.__name__, attr))
        self.verify()

    def verify(self):
        pass

    def printAction(self, msg):
        Util.printAction(msg)

    def printStep(self, msg):
        Util.printStep(msg)
    
    def printDetail(self, msg, verboseThreshold=Util.VERBOSE_LEVEL_NORMAL):
        Util.printDetail(msg, self.verboseLevel, verboseThreshold)

    def printWarning(self, msg):
        Util.printWarning(msg)

    def printError(self, msg):
        Util.printError(msg, exit=False)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from os.path import join, expanduser

userHome = expanduser('~')

sshPublicKeyLocation = join(userHome, '.ssh/id_rsa.pub')

_globusDirectory = join(userHome, '.globus')

pemDirectory = _globusDirectory
pemCertificateFile = 'usercert.pem'
pemKeyFile = 'userkey.pem'
pemCertificateLocation = join(pemDirectory, pemCertificateFile)
pemKeyLocation = join(pemDirectory, pemKeyFile)

p12Directory = _globusDirectory
p12CertificateFile = 'usercert.p12'
p12CertificateLocation = join(p12Directory, p12CertificateFile)

marketplaceEndpoint = 'https://marketplace.stratuslab.eu'
marketplaceProtocol = 'https'
marketplaceHostname = 'marketplace.stratuslab.eu'
marketplacePort = 80

pdiskPort = 8445
pdiskProtocol = 'https'

portTranslation = False
patServiceDbname = '/var/lib/one/ports.db'
patPortsRange = '15000:17000'
patTranslatedPorts = [22]
patNetworks = ['local']
patFirewallChainPrefix = 'PAT'
patMaxTranslations = 0
patVerboseLevel = 0

ETC_DIR = '/etc/stratuslab'
SHARE_DIR = '/var/share/stratuslab'
TEMPLATE_DIR = join(SHARE_DIR, 'template')
RESOURCES_DIR = join(SHARE_DIR, 'resources')
ONE_PROXY_DIR = '/opt/stratuslab/one-proxy'

AUTHN_CONFIG_FILE = ETC_DIR + '/authn/login-pswd.properties'

SHARE_TYPE = 'stratuslab'

CLOUD_CONF_DIR = '/etc/one/'
CLOUD_CONF_FILE = CLOUD_CONF_DIR + 'oned.conf'

CLOUD_VAR_LIB_DIR = '/var/lib/one'




#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import Util
import hashlib
import gzip
import bz2
from Exceptions import ExecutionException

class Compressor(object):

    compressionFormats = ('gz', 'bz2')

    @staticmethod
    def compress(filename):
        pass
        
    @staticmethod        
    def inflate(filename):
        cmd = Compressor._getInflateCommand(filename)
    
        ret = Util.execute([cmd, filename])
        if ret != 0:
            raise ExecutionException('Error inflating file: %s' % filename)

    @staticmethod
    def getCompressionFormat(filename):
        """If the file suffix (ignoring case) is in the list of supported
           compression formats, this returns the compression format.
           Otherwise, it returns the empty string. """
        suffix = Util.fileGetExtension(filename).lower()
        if suffix in Compressor.compressionFormats:
            return suffix
        else:
            return ''

    @staticmethod
    def openCompressedFile(filename, options='rb'):
        """Returns an open file handle for the given filename.  If the
           filename ends with a gzip or bzip2 suffix, then the file is
           opened as a gzip or bzip2 file.  Otherwise it is opened
           without any compression filter."""
        type = Compressor.getCompressionFormat(filename)
        if (type == 'gz'):
            return gzip.open(filename, options)
        elif (type == 'bz2'):
            return bz2.BZ2File(filename, options)
        else:
            return open(filename, options)

    @staticmethod
    def checksum_file(filename, checksums=[], chunk_size=1024*1024*10):
        """Return dictionary of checksums."""

        return Compressor._checksum_f(Compressor.openCompressedFile(filename, 'rb'), checksums, chunk_size)

    @staticmethod
    def _checksum_f(f, checksums=[], chunk_size=1024*1024*10):
        """Returns a tuple with the file size in bytes and a dictionary of
        checksums. The file named by the file handle will be fully read
        if checksums are requested.  This method will close the file
        handle."""
        
        #
        # "with" cannot be used here because the gzip library in python
        # 2.6 doesn't support the __exit__ attribute needed for it
        #
        try: 

            if not checksums:
                return {}

            digesters = []
            try:
                digesters = map(hashlib.new, checksums)
            except ValueError as e:
                raise ExecutionException('%s' % e)

            bytes = 0
            for chunk in iter((lambda:f.read(chunk_size)),''):
                bytes += len(chunk)
                for digester in digesters:
                    digester.update(chunk)

            digests = [d.hexdigest() for d in digesters]

            return (bytes, dict(zip(checksums, digests)))

        finally:
            f.close()

    @staticmethod        
    def _getCompressionCommand(filename):
        format = filename.split('.')[-1]
        return Compressor._getCompressionCommandByFormat(format)

    @staticmethod
    def _getCompressionCommandByFormat(format):
        if format == 'gz':
            cmd = 'gzip'
        elif format == 'bz2':
            cmd = 'bzip2'
        else:
            raise NotImplementedError('Unknown compression format: %s' % format)

        return cmd

    @staticmethod        
    def _getInflateCommand(filename):
        format = filename.split('.')[-1]

        if format == 'gz':
            cmd = 'gunzip'
        elif format == 'bz2':
            cmd = 'bunzip2'
        else:
            raise NotImplementedError('Unknown compression format: %s' % format)

        return cmd



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import shutil
from Configurable import Configurable
from Exceptions import ConfigurationException


class Configurator(Configurable):
    def __init__(self, configHolder):
        self.section = None
        super(Configurator, self).__init__(configHolder)

        self.baseConfigFile = self.configFile + '.ref'
        self.config = None
        self.baseConfig = None
        self._load()

    def _load(self):
        self.printDetail('Loading configuration file %s' % self.baseConfigFile)
        self.baseConfig = self.parseConfig(self.baseConfigFile)
        self._createConfigIfRequired()
        self.printDetail('Loading configuration file %s' % self.configFile)
        self.config = self.parseConfig(self.configFile)

    def _createConfigIfRequired(self):
        if not os.path.isfile(self.configFile):
            self.printDetail('Creating user configuration file %s' % self.configFile)
            shutil.copy(self.baseConfigFile, self.configFile)

    def getValue(self, key):
        section, _ = self._findSectionAndValue(key, self.config)
        return self.config.get(section, key)

    def formatDefaultKeys(self):
        columnSize = 25
        defaultConfig = self.convertToSectionDict(self.parseConfig(self.baseConfigFile))

        width = columnSize * 3 + 1
        line = '-' * width
        doubleLine = '=' * width
        startEmphasis = '\033[1;31m'
        stopEmphasis = '\033[0m'

        fields = {'line': line,
                  'doubleLine': doubleLine,
                  'startEmphasis': startEmphasis,
                  'stopEmphasis': stopEmphasis,
                  'section': 'Section'.center(width),
                  'first': 'Config key'.ljust(columnSize),
                  'second': 'Current value'.ljust(columnSize),
                  'third': 'Default value'}

        header = """
%(doubleLine)s
%(startEmphasis)s%(section)s%(stopEmphasis)s
%(line)

 %(startEmphasis)s%(first)s%(stopEmphasis)s|  %(startEmphasis)s%(second)s%(stopEmphasis)s|  %(startEmphasis)s%(third)s%(stopEmphasis)s

"""

        result = header % fields

        sections = defaultConfig.keys()
        if self.section:
            sections = [self.section]

        first = True
        for section in sections:
            if first:
                result += "%(doubleLine)s\n" % fields
                first = False
            else:
                result += "%(line)s\n" % fields

            result += '%s%s%s'.center(width) % (startEmphasis, section, stopEmphasis)
            result += "\n%(line)s\n" % fields

            keys = defaultConfig[section].keys()
            for key in keys.sort():
                result += ' %s|  %s|  %s' % (key.ljust(columnSize),
                                             self.getValue(key).ljust(columnSize),
                                             defaultConfig[section].get(key))

        return result

    def _convertToDict(self, config):
        dict = {}
        for section in config.sections():
            for k, v in config.items(section):
                dict[k] = v
        return dict

    def setOption(self, key, value):
        section, _ = self._findSectionAndValue(key, self.config)
        self.config.set(section, key, value)
        self.writeUserConfig()

    def _findSectionAndValue(self, key, config):
        section = None
        value = None
        for s in config.sections():
            for k, v in config.items(s):
                if k == key:
                    section = s
                    value = v
                    break
            if section:
                break
        if not section:
            raise ConfigurationException('Key %s does not exist' % key)
        return section, value

    def writeUserConfig(self):
        fd = open(self.configFile, 'wb')
        self.config.write(fd)
        fd.close()

    def revertConfig(self):
        if os.path.isfile(self.userConfigFile):
            os.remove(self.userConfigFile)


class SimpleConfigParser(object):
    FILENAME = 'Unknown'

    def __init__(self):
        self.items = {}

    def load(self, file=None):

        if not file:
            file = open(SimpleConfigParser.FILENAME)

        lines = file.read().split('\n')
        lines = filter(None, lines)
        lines = filter(lambda x: not x.strip().startswith('#'), lines)
        for line in lines:
            parts = line.split('=')
            if len(parts) <= 1:
                raise ValueError('Invalid configuration file format')
            username = parts[0]
            self.items[username] = self.parse_value(parts[1])

    def parse_value(self, value):
        return value

    def get(self, username):
        return self.items[username]



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, Centre National de la Recherche Scientifique (CNRS)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest
import httplib2
import urllib

import ldap

from subprocess import call

class LdapAuthenticationTest(unittest.TestCase):

    oneEndpoint = 'onehost-5.lal.in2p3.fr'
    pemCert = '/tmp/stratuslab-test-user/stratuslab-test-user-cert.pem'
    pemKey = '/tmp/stratuslab-test-user/stratuslab-test-user-key.pem'
    baseUrl = 'https://onehost-5.lal.in2p3.fr:8444'
    ldapUrl = 'ldap://onehost-5.lal.in2p3.fr:389/'

    managerDn = 'cn=admin,o=cloud'
    managerPassword = 'secret'

    username = 'jsmith'
    userDn = 'uid=jsmith,ou=users,o=cloud'
    userCertDn = 'CN=John Smith,OU=Test,O=StratusLab,C=EU'
    userPassword = 'dummy_password'

    cloudAccessDn = 'cn=cloud-access,ou=groups,o=cloud'
    
    headers = {'cache-control':'no-cache'}

    def registerUser(self):
        h = httplib2.Http(disable_ssl_certificate_validation=True)
        h.follow_redirects = False

        data = {
            'uid':self.username, 
            'mail':'smith@example.org',
            'givenName':'john',
            'sn':'smith',
            'seeAlso':self.userCertDn,
            'newUserPassword':self.userPassword,
            'newUserPasswordCheck':self.userPassword,
            'message':'I love clouds',
            'agreement':'true'
            }

        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        body = urllib.urlencode(data)
        
        url = self.baseUrl + '/users/'
        resp, content = h.request(url, "POST", body=body, headers=headers)

        # A correct registration should return a redirect. 
        self.assertEquals(303, resp.status, url + ': invalid status ' + str(resp.status))


    def verifyProfile(self):
        h = httplib2.Http(disable_ssl_certificate_validation=True)
        h.add_credentials(self.username, self.userPassword)
        url = self.baseUrl + '/profile/'
        resp, content = h.request(url)

        self.assertEquals(200, resp.status, url + ': invalid status ' + str(resp.status))

        length = len(content)
        self.assertTrue(length > 0, url + ': invalid length ' + str(length))


    def getCloudAccessGroup(self):
        ldapObj = ldap.initialize(self.ldapUrl)
        ldapObj.simple_bind_s(self.managerDn, self.managerPassword)
        
        searchScope = ldap.SCOPE_BASE
        searchFilter = '(&(objectClass=groupOfUniqueNames)(cn=cloud-access))'
        retrieveAttributes = [ 'uniqueMember' ]

        resultData = ldapObj.search_s(self.cloudAccessDn, searchScope, 
                                      searchFilter, retrieveAttributes)

        self.assertTrue(len(resultData) > 0, "no cloud-access group found encountered")
        self.assertTrue(len(resultData[0]) > 0, "malformed tuple for group information")

        return resultData[0][1]


    def authorizeUser(self):
        ldapObj = ldap.initialize(self.ldapUrl)
        ldapObj.simple_bind_s(self.managerDn, self.managerPassword)

        modlist = [(ldap.MOD_ADD, 'uniquemember', self.userDn)]

        ldapObj.modify_s(self.cloudAccessDn, modlist)


    def unauthorizeUser(self):
        ldapObj = ldap.initialize(self.ldapUrl)
        ldapObj.simple_bind_s(self.managerDn, self.managerPassword)

        modlist = [(ldap.MOD_DELETE, 'uniquemember', self.userDn)]

        ldapObj.modify_s(self.cloudAccessDn, modlist)


    def deleteUserFromLdap(self):
        ldapObj = ldap.initialize(self.ldapUrl)
        ldapObj.simple_bind_s(self.managerDn, self.managerPassword)
        
        ldapObj.delete_s(self.userDn)


    def tryUsernamePassword(self):
        cmd = [ 'stratus-describe-instance', 
                '--endpoint', self.oneEndpoint, 
                '-u', self.username, 
                '-p', self.userPassword ]

        return call(cmd)


    def tryCertificate(self):
        cmd = [ 'stratus-describe-instance', 
                '--endpoint', self.oneEndpoint, 
                '--pem-cert', self.pemCert, 
                '--pem-key', self.pemKey ]

        return call(cmd)


    def testLdapAuthentication(self):

        try:
            
            rc = self.tryUsernamePassword()
            self.assertTrue(rc != 0, 'unregistered user accessed service with username password')

            rc = self.tryCertificate()
            self.assertTrue(rc != 0, 'unregistered user accessed service with certificate')

            self.registerUser()
            self.verifyProfile()

            rc = self.tryUsernamePassword()
            self.assertTrue(rc != 0, 'user without role accessed service with username password')

            rc = self.tryCertificate()
            self.assertTrue(rc != 0, 'user without role accessed service with certificate')

            self.authorizeUser()

            rc = self.tryUsernamePassword()
            self.assertTrue(rc == 0, 'registered user with role could not access service with username password')
            
            rc = self.tryCertificate()
            self.assertTrue(rc == 0, 'registered user with role could not access service with certificate')
            
        finally:
            self.unauthorizeUser()
            self.deleteUserFromLdap()


if __name__ == "__main__":
    unittest.main()




#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import re
import shutil
import time
import tempfile
from datetime import datetime

from stratuslab import Exceptions
from stratuslab import Defaults
from stratuslab.Util import appendOrReplaceInFile, execute, fileAppendContent, \
    fileGetContent, filePutContent, scp, sshCmd
import stratuslab.Util as Util
from stratuslab.system.PackageInfo import PackageInfo
from stratuslab.system import Systems
from stratuslab.Exceptions import ExecutionException

class BaseSystem(object):

    os = ''
    caRepoName = 'CAs'

    def __init__(self):
        dateNow = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
        tmpdir = tempfile.gettempdir()
        self.stdout = open(os.path.join(tmpdir,('stratuslab_%s.log' % dateNow)), 'a')
        self.stderr = open(os.path.join(tmpdir,('stratuslab_%s.err' % dateNow)), 'a')

        self.extraRepos = {}
        self.packages = {}
        self.installPackagesErrorMsgs = []
        self.repoFileNamePattern = '/etc/%s'
        self.certificateAuthorityPackages = ''
        self.certificateAuthorityRepo = ''

        self.workOnFrontend()
        self.oneDbUsername = None
        self.oneDbPassword = None

        self.qemuConf = '/etc/libvirt/qemu.conf'

        self.shareType = Defaults.SHARE_TYPE

    def init(self):
        self._setOneHome()

    # -------------------------------------------
    #     Packages manager and related
    # -------------------------------------------

    def addRepositories(self, packages):
        pass

    def updatePackageManager(self):
        pass

    def installWebServer(self):
        pass

    def installPackages(self, packages):
        if len(packages) < 1:
            return

        self.addRepositories(packages)

        packages_versioned = []
        for package in packages:
            packages_versioned.append(
                            self.getPackageWithVersionForInstall(package))

        cmd = '%s %s' % (self.installCmd, ' '.join(packages_versioned))
        rc, output = self._executeWithOutput(cmd, shell=True)
        if rc != 0:
            raise ExecutionException('Failed to install: %s\n%s' % \
                                     (', '.join(packages_versioned), output))

        Util.printDetail(output, self.verboseLevel, Util.VERBOSE_LEVEL_DETAILED)

    def getPackageWithVersionForInstall(self, package):
        try:
            self.packages[package]
        except KeyError:
            return package
        else:
            if self.packages[package].packageVersion:
                return '%s%s%s*' % (self.packages[package].packageName,
                                    self._getPackageAndVersionSeparator(),
                                    self.packages[package].packageVersion)
            else:
                return self.packages[package].packageName

    def _getPackageAndVersionSeparator(self):
        return Systems.getPackageAndVersionSeparatorBasedOnOs(self.os)


    def installNodePackages(self, packages):
        if len(packages) > 0:
            rc, output = self._nodeShell('%s %s' %
                                     (self.installCmd, ' '.join(packages)),
                                     withOutput=True)
            if rc != 0:
                raise Exceptions.ExecutionException('Error installing packages: %s\n%s' % \
                                         (packages, output))
            for err in self.installPackagesErrorMsgs:
                if re.search(err, output, re.M):
                    raise Exceptions.ExecutionException('Error installing packages: %s\n%s' % \
                                             (packages, output))

    def installFrontendDependencies(self):
        self.addRepositories(self.frontendDeps)
        self.updatePackageManager()
        self.installPackages(self.frontendDeps)

    def installNodeDependencies(self):
        self.installNodePackages(self.nodeDeps)

    def installHypervisor(self):
        self.installNodePackages(self.hypervisorDeps.get(self.hypervisor))


    def _updatePackageAndRepoInfo(self, packageName, repoName, repoConf):
        self.packages[packageName] = PackageInfo(packageName, repository=repoName)
        self.extraRepos[repoName] = {'content' : repoConf,
                                     'filename' : self.repoFileNamePattern % repoName}

    def getPackageName(self, package):
        return self.packages[package].packageName

    def getPackageConfigFileName(self, package):
        return self.packages[package].configFile

    def getPackageInitdScriptName(self, package):
        return self.packages[package].initdScriptName

    def getPakcageRepositoryName(self, package):
        return self.packages[package].repository

    def getPakcageRepositoryConfig(self, package):
        repoName = self.getPakcageRepositoryName(package)
        return self.extraRepos[repoName]

    def getIsPackageInstalledCommand(self, package):
        pass

    def isPackageInstalled(self, package):
        cmd = self.getIsPackageInstalledCommand(package)

        rc, output = self._executeWithOutput(cmd, shell=True)

        if rc != 0:
            Util.printDetail(output)
            return False

        return True

    def startService(self, service):
        return self._operationOnService(service, 'start')

    def stopService(self, service):
        return self._operationOnService(service, 'stop')

    def restartService(self, service):
        return self._operationOnService(service, 'restart')

    def _operationOnService(self, service, operation):
        cmd = ['service', service, operation]
        rc, output = self._executeWithOutput(cmd)
        if rc != 0:
            Util.printDetail(output)
        return rc

    def startCloudSystem(self):
        self.stopService('oned')
        if self.startService('oned'):
            Util.printError("ONE failed to start")
        Util.printDetail('Waiting for ONE to finish starting')
        time.sleep(10)

    def enableServiceOnBoot(self, service, level='3'):
        return 0

    # -------------------------------------------
    #     ONE admin creation
    # -------------------------------------------

    def createCloudGroup(self, groupname, gid):
        self.oneGroup = groupname
        self.oneGid = gid

        self.executeCmd(['groupadd', '-g', self.oneGid,
                        self.oneGroup])

    def createCloudAdmin(self):

        self.createDirsCmd(os.path.dirname(self.oneHome))

        self.executeCmd(['useradd', '-d', self.oneHome, '-g',
                        self.oneGroup, '-u', self.oneUid, self.oneUsername,
                        '-s', '/bin/bash', '-p', self.onePassword, '--create-home',
                        '--expiredate ""', '--inactive -1'])

    # -------------------------------------------
    #     ONE admin env config and related
    # -------------------------------------------

    def configureCloudAdminEnv(self, ONeDPort, stratuslabLocation):
        self.ONeDPort = ONeDPort

        self.appendOrReplaceInFileCmd('%s/.bashrc' % self.oneHome,
                                      'export ONE_LOCATION',
                                      'export ONE_LOCATION=%s' % self.oneHome)
        self.appendOrReplaceInFileCmd('%s/.bashrc' % self.oneHome,
                                      'export ONE_XMLRPC',
                                      'export ONE_XMLRPC=http://localhost:%s/RPC2' % self.ONeDPort)
        self.appendOrReplaceInFileCmd('%s/.bashrc' % self.oneHome,
                                      'export PATH',
                                      'export PATH=%s/bin:%s' % (self.oneHome, os.getenv('PATH')))

        if stratuslabLocation:
            self.appendOrReplaceInFileCmd('%s/.bashrc' % self.oneHome,
                                          'export STRATUSLAB_LOCATION',
                                          'export STRATUSLAB_LOCATION=%s' % stratuslabLocation)

        self.filePutContentsCmd('%s/.bash_login' % self.oneHome,
                                '[ -f ~/.bashrc ] && source ~/.bashrc')
        self.setOwnerCmd('%s/.bash_login' % self.oneHome)

        # Hack to always load .bashrc
        self.executeCmd(['sed -i \'s/\[ -z \\\"\$PS1\\\" \\] \\&\\& '
                        'return/#&/\' %s/.bashrc' % self.oneHome], shell=True)

    def configureCloudAdminSshKeys(self):
        keyFileName = '%s/.ssh/id_rsa' % self.oneHome

        if os.path.exists(keyFileName):
            Util.printDetail('Key file %s already exists, skipping this step' % keyFileName)
            return

        self.createDirsCmd(os.path.dirname(keyFileName))
        self.setOwnerCmd(os.path.dirname(keyFileName))
        self.executeCmd(['ssh-keygen -f %s -N "" -q' % keyFileName],
                        shell=True)
        self.setOwnerCmd(keyFileName)
        self.setOwnerCmd('%s.pub' % keyFileName)

        self.copyCmd('%s.pub' % keyFileName,
                     '%s/.ssh/authorized_keys' % self.oneHome)
        self.setOwnerCmd('%s/.ssh/authorized_keys' % self.oneHome)
        self._configureCloudAdminSsh()

    def configureCloudAdminSshKeysNode(self):

        self.createDirsCmd('%s/.ssh/' % self.oneHome)
        self.setOwnerCmd('%s/.ssh/' % self.oneHome)

        # FIXME: why ssh key-pair from the Frontend is pushed to the Node?
        #        ssh-keygen on the Node should be used to generate the user
        #        specific ssh key-pair on that machine.
        oneKey = fileGetContent('%s/.ssh/id_rsa' % self.oneHome)
        self._filePutContentAsOneAdmin('%s/.ssh/id_rsa' % self.oneHome, oneKey)

        oneKeyPub = fileGetContent('%s/.ssh/id_rsa.pub' % self.oneHome)
        self._filePutContentAsOneAdmin('%s/.ssh/authorized_keys' % self.oneHome,
                                       oneKeyPub)
        self.chmodCmd('%s/.ssh/id_rsa' % self.oneHome, 0600)

        self._configureCloudAdminSsh()

    def _configureCloudAdminSsh(self):
        confFile = '%s/.ssh/config' % self.oneHome

        self.appendOrReplaceInFileCmd(confFile,
                                      '^Host.*$', 'Host *')
        self.appendOrReplaceInFileCmd(confFile,
                                      '^StrictHost.*$', 'StrictHostKeyChecking no')
        self.setOwnerCmd(confFile)
        self.chmodCmd(confFile, 0600)

    def configureCloudAdminAccount(self):
        oneAuthFile = '%s/.one/one_auth' % self.oneHome
        self.appendOrReplaceInFileCmd(oneAuthFile,
                                      self.oneUsername, '%s:%s' % (self.oneUsername, self.onePassword))
        os.environ['ONE_AUTH'] = oneAuthFile

        self.addCloudAdminToExtraGroups()

        self.configureCloudAdminSudoFrontend()

    def addCloudAdminToExtraGroups(self):
        if Util.isTrueConfVal(self.persistentDisk) and self.persistentDiskStorage == 'lvm':
            self._addCloudAdminToExtraGroup(self.persistentDiskLvmDevfilesGroup)

    def _addCloudAdminToExtraGroup(self, group):
            self.executeCmd(['usermod', '-aG', group, self.oneUsername])

    def configureCloudAdminSudoFrontend(self):
        commands = ['/sbin/lvs',
                    '/var/lib/stratuslab/python/stratuslab/tm/TMMakeVFAT.py']
        self._configureCloudAdminSudo(commands)

    def configureCloudAdminSudoNode(self):
        commands = ['/bin/chmod']
        self._configureCloudAdminSudo(commands)

    def _configureCloudAdminSudo(self, commands):
        Util.printDetail("Configuring sudo rights for '%s'" % self.oneUsername)
        for cmd in commands:
            replace = '%s ALL = NOPASSWD: %s' % (self.oneUsername, cmd)
            self.appendOrReplaceInFileCmd('/etc/sudoers', '%s' % replace, replace)

    def _setOneHome(self):
        if not self.oneHome:
            self.oneHome = os.path.expanduser('~' + self.oneUsername)

    # -------------------------------------------
    #     Persistent disks
    # -------------------------------------------

    def configureCloudAdminPdiskNode(self):
        pdiskAttach = '/usr/sbin/attach-persistent-disk.sh'
        pdiskDetach = '/usr/sbin/detach-persistent-disk.sh'

        if Util.isFalseConfVal(getattr(self, 'persistentDisks', False)):
            self.executeCmd('"[ -f %(pd)s ] || { touch %(pd)s; chmod +x %(pd)s; }"' %
                            {'pd':pdiskDetach}, shell=True)
            return

        Util.printDetail("Configuring persistent disks management for "
                         "'%s' user." % self.oneUsername)

        line = 'oneadmin ALL = NOPASSWD: %s, %s' % (pdiskAttach, pdiskDetach)
        self.appendOrReplaceInFileCmd('/etc/sudoers',
                                      '^%s.*persistent-disk.*$' %
                                      self.oneUsername, line)

    # -------------------------------------------
    #     File sharing configuration
    # -------------------------------------------

    def configureNewNfsServer(self, mountPoint, networkAddr, networkMask):
        self.createDirsCmd(mountPoint)
        self.appendOrReplaceInFileCmd('/etc/exports', '%s .*' % mountPoint,
                                      '%s %s/%s(rw,async,no_subtree_check,no_root_squash)' %
                                      (mountPoint, networkAddr, networkMask))
        self.executeCmd(['exportfs', '-a'])

    def configureExistingNfsShare(self, shareLocation, mountPoint):
        self.createDirsCmd(mountPoint)
        self.appendOrReplaceInFileCmd('/etc/fstab', '%s .*' % shareLocation,
                                      '%s %s nfs soft,intr,rsize=32768,wsize=32768,rw 0 0' % (
                                      shareLocation, mountPoint))
        self.executeCmd(['mount', '-a'])

    def configureSshServer(self):
        pass

    def configureSshClient(self, sharedDir):
        self.createDirsCmd(sharedDir)
        self.setOwnerCmd(sharedDir)

    # -------------------------------------------
    #     Hypervisor configuration
    # -------------------------------------------

    def configureHypervisor(self):
        if self.hypervisor == 'xen':
            self._configureXen()
        elif self.hypervisor == 'kvm':
            self._configureKvm()

    def _configureKvm(self):
        self.executeCmd(['modprobe', 'kvm_intel'])
        self.executeCmd(['modprobe', 'kvm_amd'])

        # seen a case when permission of /dev/kvm were 0600
        self.executeCmd(['chmod', '0666', '/dev/kvm'])

        if self.shareType == 'nfs':
                self._configureQemuUserOnFrontend()

    def _configureQemuUserOnFrontend(self):
        """Add qemu user on Fronted with the same UID and GID as on the node
        being configured. Add qemu user to 'cloud' group both on Frontend
        and the node.
        """
        if self.shareType != 'nfs':
                return


        user = group = 'qemu'
        getUidGidCmd = "getent passwd %s"

        Util.printDetail("Configuring '%s' user on Frontend as shared filesystem setup requested." % user)

        def getUidGidFromNode(user):
            rc, output = self._nodeShell(getUidGidCmd % user,
                                         withOutput=True)
            if rc != 0:
                Util.printError("Error getting '%s' user UID/GID from Node.\n%s" %
                                    (user,output))

            return _extractUidGidFromGetentPasswdOutput(output)
        def _extractUidGidFromGetentPasswdOutput(output):
            uid, gid = output.split(':')[2:4] # uid, gid
            if not all([uid, gid]):
                Util.printError("Error extracting '%s' user UID/GID from output.\n%s" %
                                    (user,output))
            return uid, gid

        uidNode, gidNode = getUidGidFromNode(user)

        rc, output = self._executeWithOutput((getUidGidCmd % uidNode).split())
        if rc == 0:
            uidLocal, gidLocal = _extractUidGidFromGetentPasswdOutput(output)
            Util.printDetail("User with UID:%s/GID:%s already configured on Frontend." %
                             (uidLocal, gidLocal), verboseLevel=self.verboseLevel)

            if gidNode != gidLocal:
                Util.printError("Frontend user '%s' GID:%s doesn't match GID:%s on Node %s." %
                                 (gidLocal, user, gidNode, self.nodeAddr))
        else:
            self._execute(['groupadd', '-g', gidNode, '-r', group])
            self._execute(['useradd', '-r', '-u', uidNode, '-g', group,
                             '-d', '/', '-s', '/sbin/nologin',
                             '-c', '"%s user"'%user, user])

        # Instruct libvirt to run VMs with GID of ONE group.
        self.appendOrReplaceInFileCmd(self.qemuConf, '^group.*$',
                                      'group = "%s"' % self.oneGroup)

        # TODO: check why this didn't work
#        # Add the user to ONE admin group. Directory with the images on
#        # shared Frontend is restricted to ONE admin user.
#        cmd = ['usermod', '-aG', self.oneGroup, user]
#        self._execute(cmd)
#        self._nodeShell(cmd)

    def _configureXen(self):
        self.appendOrReplaceInFileCmd('/etc/sudoers', self.oneUsername,
                                      '%s  ALL=(ALL) NOPASSWD: /usr/sbin/xm *' % self.oneUsername)
        self.appendOrReplaceInFileCmd('/etc/sudoers', self.oneUsername,
                                      '%s  ALL=(ALL) NOPASSWD: /usr/sbin/xentop *' % self.oneUsername)
        self.executeCmd(['sed -i -E \'s/Defaults[[:space:]]+requiretty/#&/\''
                        ' /etc/sudoers'])

    def configureLibvirt(self):
        libvirtConf = '/etc/libvirt/libvirtd.conf'

        self.appendOrReplaceInFileCmd(libvirtConf, '^unix_sock_group.*$',
                                      'unix_sock_group = "cloud"')
        self.appendOrReplaceInFileCmd(libvirtConf, '^unix_sock_ro_perms.*$',
                                      'unix_sock_ro_perms = "0777"')
        self.appendOrReplaceInFileCmd(libvirtConf, '^unix_sock_rw_perms.*$',
                                      'unix_sock_rw_perms = "0770"')
        self.appendOrReplaceInFileCmd(libvirtConf, '^auth_unix_ro.*$',
                                      'auth_unix_ro = "none"')
        self.appendOrReplaceInFileCmd(libvirtConf, '^auth_unix_rw.*$',
                                      'auth_unix_rw = "none"')
        self.appendOrReplaceInFileCmd(self.qemuConf, '^vnc_listen.*$',
                                      'vnc_listen = "0.0.0.0"')

    def startLibvirt(self):
        rc, output = self.executeCmd('service libvirtd restart'.split(),
                                     withOutput=True)
        if rc != 0:
            Util.printError('Could not start libvirt.\n%s' % output)

    # -------------------------------------------
    #     Front-end related methods
    # -------------------------------------------

    def execute(self, commandAndArgsList, **kwargs):
        return self._execute(commandAndArgsList, **kwargs)

    def _execute(self, commandAndArgsList, **kwargs):
        stdout = kwargs.get('stdout', self.stdout)
        stderr = kwargs.get('stderr', self.stderr)

        if kwargs.has_key('stdout'):
            del kwargs['stdout']
        if kwargs.has_key('stderr'):
            del kwargs['stderr']

        return execute(commandAndArgsList,
                       stdout=stdout,
                       stderr=stderr,
                       verboseLevel=self.verboseLevel,
                       verboseThreshold=Util.VERBOSE_LEVEL_DETAILED,
                       **kwargs)

    def _executeWithOutput(self, command, **kwargs):
        kwargs['withOutput'] = True
        return self._execute(command, **kwargs)

    def _setCloudAdminOwner(self, path):
        os.chown(path, int(self.oneUid), int(self.oneGid))

    def _createDirs(self, path):
        if not os.path.isdir(path) and not os.path.isfile(path):
            os.makedirs(path)

    def _copy(self, src, dst):
        if os.path.isfile(src):
            shutil.copy(src, dst)
        else:
            shutil.copytree(src, dst)

    def _remove(self, path):
        if os.path.isfile(path):
            os.remove(path)
        else:
            shutil.rmtree(path)

    # -------------------------------------------
    #     Node related methods
    # -------------------------------------------

    def _nodeShell(self, command, **kwargs):
        stdout = kwargs.get('stdout', self.stdout)
        stderr = kwargs.get('stderr', self.stderr)

        if kwargs.has_key('stdout'):
            del kwargs['stdout']
        if kwargs.has_key('stderr'):
            del kwargs['stderr']

        if isinstance(command, list):
            command = ' '.join(command)

        return sshCmd(command,
                      self.nodeAddr,
                      sshKey=self.nodePrivateKey,
                      stdout=stdout,
                      stderr=stderr,
                      verboseLevel=self.verboseLevel,
                      verboseThreshold=Util.VERBOSE_LEVEL_DETAILED,
                      **kwargs)

    def _nodeCopy(self, source, dest, **kwargs):
        stdout = kwargs.get('stdout', self.stdout)
        stderr = kwargs.get('stderr', self.stderr)

        if kwargs.has_key('stdout'):
            del kwargs['stdout']
        if kwargs.has_key('stderr'):
            del kwargs['stderr']

        return scp(source,
                   'root@%s:%s' % (self.nodeAddr, dest),
                   self.nodePrivateKey,
                   stdout=stdout,
                   stderr=stderr,
                   verboseLevel=self.verboseLevel,
                   verboseThreshold=Util.VERBOSE_LEVEL_DETAILED,
                   **kwargs)

    def _remoteSetCloudAdminOwner(self, path):
        self._nodeShell(['chown %s:%s %s' % (self.oneUid,
                        self.oneGid, path)])

    def _remoteCreateDirs(self, path):
        self._nodeShell('mkdir -p %s' % path)

    def _remoteAppendOrReplaceInFile(self, filename, search, replace):
        res = self._nodeShell(['grep', '"%s"'%search, filename])

        replace = Util.escapeDoubleQuotes(replace)

        if self._patternExists(res):
            rc, output = self._nodeShell('"sed -i \'s|%s|%s|\' %s"' % (search, replace, filename),
                                         withOutput=True, shell=True)
            if rc != 0:
                Util.printError("Failed to modify %s.\n%s" % (filename, output))
        else:
            self._remoteFileAppendContents(filename, replace)

    def _patternExists(self, returnCode):
        return returnCode == 0

    def _remoteCopyFile(self, src, dest):
        self._nodeShell(['cp -rf %s %s' % (src, dest)])

    def _remoteRemove(self, path):
        self._nodeShell(['rm -rf %s' % path])

    def _remoteFilePutContents(self, filename, data):
        data = Util.escapeDoubleQuotes(data, times=4)

        rc, output = self._nodeShell('"echo \\"%s\\" > %s"' % (data, filename),
                                     withOutput=True, shell=True)
        if rc != 0:
            Util.printError("Failed to write to %s\n%s" % (filename, output))

    def _remoteFileAppendContents(self, filename, data):
        data = Util.escapeDoubleQuotes(data, times=4)

        rc, output = self._nodeShell('"echo \\"%s\\" >> %s"' % (data, filename),
                                     withOutput=True, shell=True)
        if rc != 0:
            Util.printError("Failed to append to %s\n%s" % (filename, output))

    def _filePutContentAsOneAdmin(self, filename, content):
        self.filePutContentsCmd(filename, content)
        self.setOwnerCmd(filename)

    def _remoteChmod(self, path, mode):
        return self._nodeShell('chmod %o %s' % (mode, path))

    def _remoteFileExists(self, path):
        return self._nodeShell('ls %s' % path, sshQuiet=True) == 0

    # -------------------------------------------
    #     General
    # -------------------------------------------

    def setNodeAddr(self, nodeAddr):
        self.nodeAddr = nodeAddr

    def setNodePrivateKey(self, privateKey):
        self.nodePrivateKey = privateKey

    def workOnFrontend(self):
        self.appendOrReplaceInFileCmd = appendOrReplaceInFile
        self.setOwnerCmd = self._setCloudAdminOwner
        self.executeCmd = self._execute
        self.executeCmdWithOutput = self._executeWithOutput
        self.createDirsCmd = self._createDirs
        self.filePutContentsCmd = filePutContent
        self.fileAppendContentsCmd = fileAppendContent
        self.chmodCmd = os.chmod
        self.copyCmd = self._copy
        self.duplicateCmd = self._copy
        self.removeCmd = self._remove

    def workOnNode(self):
        self.appendOrReplaceInFileCmd = self._remoteAppendOrReplaceInFile
        self.setOwnerCmd = self._remoteSetCloudAdminOwner
        self.executeCmd = self._nodeShell
        self.duplicateCmd = self._remoteCopyFile
        self.createDirsCmd = self._remoteCreateDirs
        self.filePutContentsCmd = self._remoteFilePutContents
        self.fileAppendContentsCmd = self._remoteFileAppendContents
        self.chmodCmd = self._remoteChmod
        self.copyCmd = self._nodeCopy
        self.removeCmd = self._remoteRemove

    def configureQuarantine(self):
        filename = os.path.join(Defaults.ETC_DIR, 'quarantine.cfg')
        search = '^PERIOD.*$'
        replace = 'PERIOD=%(quarantinePeriod)s' % self.__dict__
        Util.appendOrReplaceInFile(filename, search, replace)

    def configureCloudProxyService(self):
        self.installPackages(['stratuslab-one-proxy'])
        self._configureProxyDefaultUsers()
        self._restartJetty()

    def _configureProxyDefaultUsers(self):
        self._configureProxyDefaultUsersUsernamePassword()

    def _configureProxyDefaultUsersUsernamePassword(self):
        filename = Defaults.AUTHN_CONFIG_FILE
        search = self.oneUsername
        replace = '%(oneUsername)s=%(proxyOneadminPassword)s,cloud-access' % self.__dict__
        Util.appendOrReplaceInFile(filename, search, replace)

    def _restartJetty(self):
        self.executeCmd('/etc/init.d/one-proxy restart'.split(' '))

    # -------------------------------------------
    # Firewall
    # -------------------------------------------

    # TODO: extract Firewall class from the code below

    DEFAULT_FIREWALL_TABLE = 'filter'

    # redefine in sub-class to point to required file
    FILE_FIREWALL_RULES = '/etc/sysconfig/iptables'

    IP_TABLES_LIST = ['filter','nat','mangle','raw']

    def _configureNetworkInterface(self, device, ip, netmask):
        pass

    def configureFirewall(self):
        self._loadNetfilterModules()

        self._configureFirewallForProxy()
        self._configureFirewallNat()
        self._persistFirewallRules()

    def _configureFirewallForProxy(self):
        port = str(self.onePort)
        rules = ({'table':'filter',
                  'rule' :'-A INPUT -s 127.0.0.1 -p tcp -m tcp --dport %s -j ACCEPT' % port},
                 {'table':'filter',
                  'rule' :'-A INPUT -p tcp -m tcp --dport %s -j REJECT --reject-with icmp-port-unreachable' % port})

        if not self._isSetFirewallRulesAll(rules):
            self._setFirewallRulesAll(rules)

    def _configureFirewallNat(self):
        if self.nat.lower() in ['false', 'no', 'off', '0', '']:
            return None

        self._configureFirewallNatNetworking()

        networkWithMask = '%s/%s' % (self.natNetwork, self.natNetmask)
        rules = ({'table':'nat',
                  'rule':'-A POSTROUTING -s %s -d ! %s -j MASQUERADE' % ((networkWithMask,)*2)},
                  {'table':'filter',
                   'rule':'-A FORWARD -d %s -m state --state RELATED,ESTABLISHED -j ACCEPT' % networkWithMask},
                  {'table':'filter',
                   'rule':'-A FORWARD -d %s -j ACCEPT' % networkWithMask})

        if not self._isSetFirewallRulesAll(rules):
            self._setFirewallRulesAll(rules)

    def _configureFirewallNatNetworking(self):
        self._enableIpForwarding()

        device = self.natNetworkInterface
        ip = getattr(self, 'natGateway', '')
        if not ip:
            ip = Util.gatewayIpFromNetAddress(self.natNetwork)

        self._configureVirtualNetInterface(device, ip,
                                           self.natNetmask)

    @staticmethod
    def enableIpForwarding():
        Util.printDetail('Enabling packets forwarding.')
        file(FILE_IPFORWARD_HOT_ENABLE, 'w').write('1')
        appendOrReplaceInFile(FILE_IPFORWARD_PERSIST,
                              'net.ipv4.ip_forward',
                              'net.ipv4.ip_forward = 1')

    def _enableIpForwarding(self):
        return BaseSystem.enableIpForwarding()

    def _configureVirtualNetInterface(self, device, ip, netmask):
        device = device + ':privlan'

        Util.printDetail('Configuring network interface %s.' % device)
        self._configureNetworkInterface(device, ip, netmask)

        Util.printDetail('Starting network interface %s.' % device)
        self.executeCmd(['ifup', device])

    def _persistFirewallRules(self):
        self._saveFirewallRules(self.FILE_FIREWALL_RULES)


    def _loadNetfilterModules(self):
        # just in case if kernel modules were not yet loaded
        devNull = open(os.path.devnull, 'w')
        for table in self.IP_TABLES_LIST:
            cmd = 'iptables -nL -t %s' % table
            self.executeCmd(cmd.split(), stdout=devNull)
        devNull.close()

    def _saveFirewallRules(self, filename):
        # back-up
        self.executeCmd(('cp -fp %s %s.LAST'%((filename,)*2)).split(' '))

        _,output = self.executeCmdWithOutput(['iptables-save'])
        Util.printDetail('Saving firewall rules to %s.' % filename)
        filePutContent(filename, output)
        os.chmod(filename, 0600)

    def _isSetFirewallRulesAll(self, rules):
        tables = dict.fromkeys([r.get('table', self.DEFAULT_FIREWALL_TABLE)
                                                        for r in rules]).keys()
        currentRules = self._getFirewallRulesPerTable(tables)
        for ruleSpec in rules:
            if not self._isSetFirewallRule(currentRules, ruleSpec):
                return False
        return True

    def _getFirewallRulesPerTable(self, tables=IP_TABLES_LIST):
        rules = {}
        for table in tables:
            rc, output = self.executeCmdWithOutput(('iptables-save -t %s' %
                                                   table).split(' '))
            if rc != 0:
                raise Exceptions.ExecutionException('iptables-save reported an error:\n%s'%
                                         output)
            rules.update({table:output})
        return rules

    def _isSetFirewallRule(self, currentRules, ruleSpec):
        rule, table = self._getRuleAndTableFromRuleSpec(ruleSpec)
        rulesInTable = currentRules[table]

        if re.search(rule, rulesInTable, re.M):
            return True
        return False

    def _setFirewallRulesAll(self, rules):
        self._deleteFirewallRulesAllGiven(rules)

        for ruleSpec in rules:
            self._setFirewallRule(ruleSpec)

    def _deleteFirewallRulesAllGiven(self, rules):
        for ruleSpec in rules:
            self._deleteFirewallRule(ruleSpec)

    def _deleteFirewallRule(self, ruleSpec):
        rule, table = self._getRuleAndTableFromRuleSpec(ruleSpec)
        rule = '-D %s' % rule[3:] # remove action; leave chain and rule

        self.executeCmd(('iptables -t %s %s' % (table,rule)).split(' '))

    def _setFirewallRule(self, ruleSpec):
        rule, table = self._getRuleAndTableFromRuleSpec(ruleSpec)

        self.executeCmd(('iptables -t %s %s' % (table,rule)).split(' '))

    def _getRuleAndTableFromRuleSpec(self, ruleSpec):
        return ruleSpec['rule'], \
               ruleSpec.get('table', self.DEFAULT_FIREWALL_TABLE)

    # -------------------------------------------
    # sendmail
    # -------------------------------------------

    def installSendmail(self):
        package = 'sendmail'
        if getattr(self, 'smtpHost', 'localhost') == 'localhost':
            self.installPackages([self.getPackageName(package)])
        else:
            Util.printDetail('Skipping installation of %s' % package)

    # -------------------------------------------
    # CA
    # -------------------------------------------

    def installCAs(self):
        def _isCertificateAuthority():
            return Util.isTrueConfVal(getattr(self, 'certificateAuthority', False))

        if not _isCertificateAuthority():
            Util.printDetail('Requested not to install CAs.')
        else:
            self._installCAs()
            self._installFetchCrl()
            self._enableFetchCrl()

    def _enableFetchCrl(self):
        pass

    def _installCAs(self):
        packages = []

        if self.certificateAuthorityPackages and self.certificateAuthorityRepo:
            caPackages = map(lambda x: x.strip(),
                             self.certificateAuthorityPackages.split(','))
            packages.extend(caPackages)
            repoConf = '\n'.join([line.strip() for line in self.certificateAuthorityRepo.strip().split('\n')])
            repoName = self.caRepoName
            for package in packages:
                self._updatePackageAndRepoInfo(package, repoName, repoConf)
        else:
            packages.append(self.getPackageName('CA'))

        self.installPackages(packages)

        for package in packages:
            if not self.isPackageInstalled(package):
                Util.printError('Failed to install %s.' % package)

    def _installFetchCrl(self):
        package = self.getPackageName('fetch-crl')
        self.installPackages([package])
        if not self.isPackageInstalled(package):
            Util.printError('Failed to install %s.' % package)

    # -------------------------------------------
    # DHCP server
    # -------------------------------------------

    NET_TYPES_DHCP = ['OnePublicNetwork', 'OneLocalNetwork']

    def configureDhcpServer(self):

        def _dhcpDefined():
            return Util.isTrueConfVal(getattr(self, 'dhcp', 'False'))
        def _dhcpNetTypesDefined():
            return any([Util.isTrueConfVal(getattr(self, self._assembleDhcpAttributeName(v), 'False'))
                                                for v in self.NET_TYPES_DHCP])

        if not _dhcpDefined():
            return
        elif not _dhcpNetTypesDefined():
            return

        Util.printStep('Configuring DHCP service')
        self._installDhcp()
        self._confgureDhcp()
        self._startDhcp()

    def _assembleDhcpAttributeName(self, postfix):
        DHCP_PARAMETER_PREFIX = 'dhcp'
        return '%s%s' % (DHCP_PARAMETER_PREFIX, postfix)

    def _installDhcp(self):
        Util.printDetail('Installing DHCP server.')

        dhcpPackage = self.getPackageName('dhcp')
        self.installPackages([dhcpPackage])

        if not self.isPackageInstalled(dhcpPackage):
            Util.printError('Failed to install %s.' % dhcpPackage)

    def _confgureDhcp(self):

        def _isAllDhcpGroupsDefined(_groups):
            return all(_groups.values())

        def _getConfGlobals():
            _globals = """
ddns-update-style none;
ignore unknown-clients;
ignore bootp;
"""
            if hasattr(self, 'dhcpNtpServers') and self.dhcpNtpServers:
                _globals += 'option ntp-servers %s;\n' % self.dhcpNtpServers
            return _globals

        def _getConfSubnets():
            subnetTemplate = """
subnet %(subnet)s netmask %(netmask)s {
  option routers %(routers)s;
}
"""
            subnet = ''
            # All net types are defined together with NATing. Assuming NATing for
            # Local net type. Need to create a shared network.
            if Util.isTrueConfVal(self.nat) and _isAllDhcpGroupsDefined(dhcpGroups):
                subnet = """
shared-network StratusLab-LAN {
"""
                for _type in self.NET_TYPES_DHCP:
                    subnet += subnetTemplate % {
                               'subnet'  : getattr(self, self._assembleDhcpAttributeName('%sSubnet' % _type)),
                               'netmask' : getattr(self, self._assembleDhcpAttributeName('%sNetmask' % _type)),
                               'routers' : getattr(self, self._assembleDhcpAttributeName('%sRouters' % _type))}
                subnet += "}\n"

            elif Util.isTrueConfVal(self.nat) and dhcpGroups['OneLocalNetwork']:
                subnet = """
shared-network StratusLab-LAN {
"""
                # main interface
                subnet += """
subnet %(subnet)s netmask %(netmask)s {
}
""" % {'subnet'  : self.dhcpSubnet,
       'netmask' : self.dhcpNetmask}
                # virtual interface
                natGateway = getattr(self, 'natGateway', '')
                if not natGateway:
                    natGateway = Util.gatewayIpFromNetAddress(self.natNetwork)
                subnet += subnetTemplate % {'subnet'  : self.natNetwork,
                                            'netmask' : self.natNetmask,
                                            'routers' : natGateway}
                subnet += "}\n"

            elif dhcpGroups['OnePublicNetwork']:
                # main interface
                subnet += """
subnet %(subnet)s netmask %(netmask)s {
}
""" % {'subnet'  : self.dhcpSubnet,
       'netmask' : self.dhcpNetmask}

            elif dhcpGroups['OneLocalNetwork']:
                # virtual interface
                subnet = subnetTemplate % {
                                'subnet'  : self.dhcpOneLocalNetworkSubnet,
                                'netmask' : self.dhcpOneLocalNetworkNetmask,
                                'routers' : self.dhcpOneLocalNetworkRouters}
            else:
                Util.printWarning('Invalid parameters combination to configure DHCP.')

            return subnet

        def _getConfGroups():
            groupHeadTemplate = """
group {
  option broadcast-address %(broadcast)s;
  option subnet-mask %(netmask)s;
  option routers %(routers)s;
  option domain-name "%(domainName)s";
  option domain-name-servers %(nameservers)s;
"""
            hostTemplate = """
  host %(type)s-vm%(id)s {
    hardware ethernet %(mac)s;
    fixed-address %(ip)s;
    max-lease-time %(leaseTime)s;
  }
"""
            groups = ''
            for _type,ipsMacs in dhcpGroups.items():
                if not ipsMacs:
                    continue
                groups += groupHeadTemplate % \
                    {'broadcast'   : getattr(self, self._assembleDhcpAttributeName('%sBroadcast' % _type)),
                     'netmask'     : getattr(self, self._assembleDhcpAttributeName('%sNetmask' % _type)),
                     'routers'     : getattr(self, self._assembleDhcpAttributeName('%sRouters' % _type)),
                     'domainName'  : getattr(self, self._assembleDhcpAttributeName('%sDomainName' % _type)),
                     'nameservers' : getattr(self, self._assembleDhcpAttributeName('%sDomainNameServers' % _type))}

                hosts = ''
                for i,ipMac in enumerate(ipsMacs):
                    hosts += hostTemplate % {'type' : _type.lower(),
                                             'id'  : str(i),
                                             'mac' : ipMac[1],
                                             'ip' : ipMac[0],
                                             'leaseTime' : self.dhcpLeaseTime}
                groups += hosts
                groups += '}\n'

            return groups

        Util.printDetail('Configuring DHCP server.')

        _NOTHING = []
        dhcpGroups = dict.fromkeys(self.NET_TYPES_DHCP, _NOTHING)

        for netType in self.NET_TYPES_DHCP:
            if Util.isTrueConfVal(getattr(self, self._assembleDhcpAttributeName(netType), False)):
                dhcpGroups[netType] = self.__getIpMacTuplesForNetworkType(netType)

        if not any(dhcpGroups.values()):
            Util.printError('When configuring DHCP %s networks IP/MAC pairs should be given.' %
                                ','.join(self.NET_TYPES_DHCP))

        content = _getConfGlobals() + \
                    _getConfSubnets() + \
                    _getConfGroups()

        confFile = self.getPackageConfigFileName('dhcp')

        Util.filePutContent(confFile, content)

    def __getIpMacTuplesForNetworkType(self, _type):
        if _type not in self.NET_TYPES_DHCP:
            Util.printError('Expected one of: %s. Got %s'%(','.join(self.NET_TYPES_DHCP),_type))

        _type = _type.replace(_type[0], _type[0].lower(), 1)

        ips = [x for x in getattr(self, '%sAddr'%_type).split()]
        macs = [x for x in getattr(self, '%sMac'%_type).split()]
        if len(ips) != len(macs):
            Util.printError('%s network: number of IPs should match number of MACs.'%_type)
        return zip(ips, macs)

    def _startDhcp(self):
        Util.printDetail('(Re)Starting DHCP server.')

        serviceName = self.packages['dhcp'].initdScriptName
        rc = self.restartService(serviceName)

        if rc != 0:
            Util.printError('Filed to (re)start DHCP service.')

    # -------------------------------------------
    # DB
    # -------------------------------------------

    def configureDatabase(self):

        if self.oneDbHost in ['localhost', '127.0.0.1']:
            Util.printDetail('Installing MySQL server.')
            mysqlPackage = self.getPackageName('MySQLServer')
            self.installPackages([mysqlPackage])

            Util.printDetail('Starting MySQL server.')
            mysqlService = self.getPackageInitdScriptName('MySQLServer')
            self.startService(mysqlService)

            Util.printDetail('Changing db root password')
            self._configureRootDbUser(self.oneDbRootPassword)

            Util.printDetail('Creating oneadmin db account')
            self._configureDbUser(self.oneDbUsername, self.oneDbPassword)
        else:
            Util.printDetail('Skipping MySQL installation/configuration. It is assumed to be configured on %s' % self.oneDbHost)

    def _configureRootDbUser(self, password):
        rc, output = self._execute(["/usr/bin/mysqladmin", "-uroot", "password", "%s" % password], withOutput=True)
        if rc != 0:
            Util.printWarning("Couldn't set root password. Already set?\n%s" % output)

    def _configureDbUser(self, username, password):
        mysqlCommand = "/usr/bin/mysql -uroot -p%s" % self.oneDbRootPassword
        userCreate = "CREATE USER '%s'@'localhost' IDENTIFIED BY '%s'" % (username, password)
        userGrant =  "GRANT CREATE, DROP, SELECT, INSERT, DELETE, UPDATE, INDEX ON opennebula.* TO '%s'@'localhost'" % username

        rc, output = self._execute("%s -e \"%s\"" % (mysqlCommand, userCreate),
                                   withOutput=True, shell=True)
        if rc != 0:
            Util.printWarning("Couldn't create user '%s'. Already exists?\n%s" % (username, output))

        rc, output = self._execute("%s -e \"%s\"" % (mysqlCommand, userGrant),
                                   withOutput=True, shell=True)
        if rc != 0:
            Util.printError("Error granting permission for user '%s'.\n%s" % (username, output))

    # -------------------------------------------
    # Bridge
    # -------------------------------------------

    def configureBridgeRemotely(self):
        def doNotConfigureBridge():
            return Util.isFalseConfVal(getattr(self, 'nodeBridgeConfigure', True))

        if doNotConfigureBridge():
            Util.printDetail('Asked not to configure bridge')
            return

        checkBridgeCmd = '"brctl show | grep ^%s.*%s$"' % \
                            (self.nodeBridgeName, self.nodeNetworkInterface)
        rc, output = self._nodeShell(checkBridgeCmd, withOutput=True, shell=True)
        if rc == 0:
            Util.printDetail('Bridge already configured')
            return
        else:
            Util.printDetail('Bridge is NOT configured. %s' % output)

        configureBridgeCmd = 'nohup "brctl addbr %(bridge)s; sleep 10; ifconfig %(interf)s 0.0.0.0; sleep 10; brctl addif %(bridge)s %(interf)s; sleep 10; dhclient %(bridge)s"' % \
                            {'bridge' : self.nodeBridgeName,
                             'interf' : self.nodeNetworkInterface}

        rc, output = self._nodeShell(configureBridgeCmd, withOutput=True, shell=True)
        if rc != 0:
            Util.printDetail('Failed to configure bridge.\n%s' % output)
        else:
            sleepTime = 5
            Util.printDetail('Sleeping %i sec for the bridge one the node to come up.' % sleepTime)
            time.sleep(sleepTime)

            Util.printDetail('Testing connection to the node.')
            rc, output = self._nodeShell('true', withOutput=True)
            if rc == 0:
                Util.printDetail('OK.')
            else:
                Util.printError('Could not connect to the node after attempt to configre bridge.\n%s' % output)

            Util.printDetail('Testing if bridge was configured.')
            rc, output = self._nodeShell(checkBridgeCmd, withOutput=True, shell=True)
            if rc == 0:
                Util.printDetail('OK.')
                self._persistRemoteBridgeConfig(self.nodeNetworkInterface, self.nodeBridgeName)
                return
            else:
                Util.printError('Bridge was not configured.\n%s' % output)

    def _persistRemoteBridgeConfig(self):
        pass

    def _writeToFilesRemote(self, listOfFileNameContentTuples):
        tmpFilename = tempfile.mktemp()
        for remoteFilename, content in listOfFileNameContentTuples:
            Util.filePutContent(tmpFilename, content)
            self._nodeCopy(tmpFilename, remoteFilename)
        try:
            os.unlink(tmpFilename)
        except: pass

FILE_IPFORWARD_HOT_ENABLE = '/proc/sys/net/ipv4/ip_forward'
FILE_IPFORWARD_PERSIST = '/etc/sysctl.conf'

def enableIpForwarding():
    return BaseSystem.enableIpForwarding()



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552.
#
# Copyright (c) 2012, Centre National de la Recherche Scientifique
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import sqlite3

from stratuslab import Defaults
from stratuslab import Util
from stratuslab.HttpClient import HttpClient
from stratuslab.pat.Core import PortTranslation, VmPortTranslation

class PortTranslationWebClient(PortTranslation):

    def __init__(self, configHolder):
        self.client = HttpClient(configHolder)
        self.patServiceHost = None
        super(PortTranslationWebClient, self).__init__(configHolder)

    def getAllPortTranslation(self):
        return self.getPortTranslation()

    def getVmPortTranslation(self, vmId):
        ports = self.getPortTranslation(vmId)
        return ports.get(vmId, VmPortTranslation(vmid=vmId))

    def getPortTranslation(self, vmId=None):
        url = self._buildUrl(vmId)
        ports = self._getAndCleanJson(url)
        return ports

    def _buildUrl(self, vmId=None):
        url = "http://%s/cgi-bin/vmports.py" % self.patServiceHost
        if vmId:
            url = url + "?vmid=%s" % vmId
        return url

    def _getAndCleanJson(self, url):
        try:
            headers, content = self._getJson(url)
            ports = json.loads(content, object_hook=VmPortTranslation.fromDict)
        except Exception, e:
            self._handleException(e)
        return ports

    def _getJson(self, url):
        headers, content = self.client.get(url, accept='application/json')
        return headers, content.replace('\\', '')

    def _handleException(self, exception):
        if self.verboseLevel >= Util.VERBOSE_LEVEL_NORMAL:
            raise exception
        else:
            raise Exception("Couldn't get port translation. Increase verbose level to see details.")


class PortTranslationDbClient(PortTranslation):
    def __init__(self, configHolder):
        super(PortTranslationDbClient, self).__init__(configHolder)

        self.conn = sqlite3.connect(self.patServiceDbname or Defaults.patServiceDbname)
        self.conn.row_factory = sqlite3.Row
        self.cursor = self.conn.cursor()

    def getAllPortTranslation(self):
        return self.getPortTranslation()

    def getVmPortTranslation(self, vmId):
        ports = self.getPortTranslation(vmId)
        return ports.get(vmId, VmPortTranslation(vmid=vmId))

    def getPortTranslation(self, vmId=None):
        query = self._buildQuery(vmId)
        rawPorts = self._executeQuery(query)
        return VmPortTranslation.fromTuple(rawPorts)

    def getAllGwPorts(self):
        query = "SELECT DISTINCT(local) FROM ports"
        rawPorts = self._executeQuery(query)
        formatedPorts = [port for port, in rawPorts]
        return formatedPorts

    def getGwPortFromVmPort(self, vmId, vmPort):
        query = "SELECT * FROM ports WHERE id = '%s' AND remote = '%s'" % (vmId, vmPort)
        return self._fetchSingleValue(query, 'local')

    def getVmPortFromGwPort(self, vmId, gwPort):
        query = "SELECT * FROM ports WHERE id = '%s' AND local = '%s'" % (vmId, gwPort)
        return self._fetchSingleValue(query, 'remote')

    def _buildQuery(self, vmId):
        query = "SELECT id, remote, local FROM ports"
        if vmId:
            query = query + " WHERE id = '%d'" % int(vmId)
        return query

    def _executeQuery(self, query):
        self.cursor.execute(query)
        rawData = self.cursor.fetchall()
        return rawData

    def _fetchSingleValue(self, query, field):
        self.cursor.execute(query)
        row = self.cursor.fetchone()
        if not row:
            return None
        return row[field]




#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2010, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import commands
import tempfile
import shutil

from stratuslab.Exceptions import ExecutionException
import stratuslab.Util as Util
from stratuslab.ConfigHolder import ConfigHolder


class CertGenerator(object):
    descriptionP12 = 'generate p12 self-signed certificate.'

    def __init__(self, configHolder=ConfigHolder()):
        self.configHolder = configHolder
        self.tmp_dir = ''

    @staticmethod
    def defaultRunOptionsP12():
        return {'commonName': 'Jane Tester',
                'outputFile': 'cert.p12',
                'certPassword': 'XYZXYZ',
                'certValidity': '365',
                'subjectEmail': 'jane.tester@example.org'}

    @staticmethod
    def addCertGeneratorOptions(parser):
        defaults = CertGenerator.defaultRunOptionsP12()

        parser.add_option('-n', '--common-name', dest='commonName',
                          help='Common Name for the certificate. (Default: %s)' %
                               defaults['commonName'],
                          default=defaults['commonName'], metavar='NAME')

        parser.add_option('-o', '--output', dest='outputFile',
                          help='file to store the generated certificate (Default: %s)' %
                               defaults['outputFile'],
                          default=defaults['outputFile'], metavar='FILE')

        parser.add_option('-p', '--password', dest='certPassword',
                          help='password for the certificate. (Default: %s)' %
                               defaults['certPassword'],
                          default=defaults['certPassword'], metavar='PASS')

        parser.add_option('--validity', dest='certValidity',
                          help='validity of the certificate. (Default: %s)' %
                               defaults['certValidity'],
                          default=defaults['certValidity'], metavar='DAYS')

        parser.add_option('-e', '--email', dest='subjectEmail',
                          help='subject email. (Default: %s)' %
                               defaults['subjectEmail'],
                          default=defaults['subjectEmail'], metavar='EMAIL')

        parser.add_option('--no-cleanup', dest='noCleanup',
                          help='do not do cleanup.',
                          default=False, action='store_true')

    @staticmethod
    def buildCertGeneratorP12Parser(parser):
        parser.description = CertGenerator.descriptionP12
        CertGenerator.addCertGeneratorOptions(parser)

    def generateP12(self):
        self.tmp_dir = tempfile.mkdtemp()
        Util.printDetail('Temporary directory for certificate generation: %s' %
                         self.tmp_dir, self.configHolder.verboseLevel,
                         Util.VERBOSE_LEVEL_DETAILED)
        try:
            self._generateOpensslConfig()
            self._runCommandsP12()
        finally:
            if self.configHolder.noCleanup:
                Util.printInfo('Intermediate files are in %s' % self.tmp_dir)
            else:
                try:
                    shutil.rmtree(self.tmp_dir, ignore_errors=True)
                except:
                    pass

    def _generateOpensslConfig(self):
        config = """
[ req ]
distinguished_name     = req_distinguished_name
x509_extensions        = v3_ca
prompt                 = no
input_password         = %(certPassword)s
output_password        = %(certPassword)s

dirstring_type = nobmp

[ req_distinguished_name ]
C = EU
O = StratusLab Project
OU = Testing Department
CN = %(commonName)s

[ v3_ca ]
basicConstraints = CA:false
nsCertType=client, email, objsign
keyUsage=critical, digitalSignature, nonRepudiation, keyEncipherment, dataEncipherment, keyAgreement
subjectKeyIdentifier=hash
authorityKeyIdentifier=keyid:always,issuer:always
subjectAltName=email:%(subjectEmail)s
""" % self.configHolder.options

        conf_filename = os.path.join(self.tmp_dir, 'openssl.cfg')
        open(conf_filename, 'w').write(config)

        Util.printDetail("Generated openssl configuration in: %s" % conf_filename,
                         self.configHolder.verboseLevel)
        Util.printDetail("Openssl configuration: %s" % open(conf_filename).read(),
                         self.configHolder.verboseLevel,
                         Util.VERBOSE_LEVEL_DETAILED)

    def _runCommandsP12(self):
        opts = dict(self.configHolder.options.items() + [('tmp', self.tmp_dir)])

        cmds = []

        cmds.append("openssl genrsa -passout pass:%(certPassword)s -des3 \
-out %(tmp)s/test-key.pem 2048" % opts)

        cmds.append("openssl req -new -key %(tmp)s/test-key.pem \
-out %(tmp)s/test.csr \
-config %(tmp)s/openssl.cfg " % opts)

        cmds.append("openssl x509 -req -days %(certValidity)s \
-in %(tmp)s/test.csr \
-signkey %(tmp)s/test-key.pem \
-out %(tmp)s/test-cert.pem \
-extfile %(tmp)s/openssl.cfg \
-extensions v3_ca \
-passin pass:%(certPassword)s" % opts)

        cmds.append("openssl pkcs12 -export -in %(tmp)s/test-cert.pem \
-inkey %(tmp)s/test-key.pem \
-out %(outputFile)s \
-passin pass:%(certPassword)s \
-passout pass:%(certPassword)s" % opts)

        for cmd in cmds:
            self._runCommand(cmd)

    def _runCommand(self, cmd):
        Util.printDetail('Running: %s' % cmd, self.configHolder.verboseLevel)
        rc, output = commands.getstatusoutput(cmd)
        if rc != 0:
            raise ExecutionException('Failed running %s:\n%s' % (cmd, output))



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from stratuslab.Runner import Runner
from stratuslab.ConfigHolder import ConfigHolder
from stratuslab.AuthnCommand import AuthnCommand


class InstanceCommand(AuthnCommand):
    """A command-line program to kill a virtual machine."""


    def __init__(self):
        self.vmIds = []
        super(InstanceCommand, self).__init__()

    def parse(self):

        self.parser.usage = self.parser_usage
        self.parser.description = self.parser_description

        self.parser.add_option('-i', '--input', dest='inVmIdsFile',
                               help='file containing line separated vm-ids', metavar='FILE',
                               default=None)

        AuthnCommand.addCloudEndpointOptions(self.parser)

        super(InstanceCommand, self).parse()

        self.options, self.vmIds = self.parser.parse_args()

    def checkOptions(self):
        AuthnCommand.checkCloudEndpointOptionsOnly(self)

        if not self.vmIds and not self.options.inVmIdsFile:
            self.parser.error('Missing vm-id or input file (-i/--input)')

        super(InstanceCommand, self).checkOptions()

    def _getRunner(self):
        configHolder = ConfigHolder(self.options.__dict__)
        return Runner(None, configHolder)

    def shutdownInstances(self):
        self._getRunner().shutdownInstances(self.vmIds)

    def killInstances(self):
        self._getRunner().killInstances(self.vmIds)




#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2012, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64
import httplib
import urllib

from stratuslab.messaging.MsgBase import MsgBase

SECRET_HEADER = {'User-Agent':'StratusLab'}

class AmazonSqsQueue(MsgBase):
    def __init__(self, configHolder):
        super(AmazonSqsQueue, self).__init__(configHolder)

        self.conn = httplib.HTTPSConnection(self.msg_endpoint)
        self.conn.debuglevel = self.verboseLevel
        self.headers = {'Accept':'*/*', 
                        'Content-Type':'application/x-www-form-urlencoded'}
        self.headers.update(SECRET_HEADER)

    @staticmethod
    def _endcode_message(message):
        return base64.b64encode(message)
    def _build_query_params(self, message):
        params = {'Action':'SendMessage',
                  'MessageBody':self._endcode_message(message)}
        return urllib.urlencode(params)

    def send(self, message):
        'message - dictionary'
        params = self._build_query_params(message)
        self.conn.request('POST', self.msg_queue, params, self.headers)
        response = self.conn.getresponse()
        status = str(response.status)
        if not status.startswith('2'):
            data = response.read()
            raise Exception(data)
        self.conn.close()



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import socket
import signal

from stratuslab.CloudConnectorFactory import CloudConnectorFactory
from stratuslab.CloudInfo import CloudInfo
from stratuslab.Configurable import Configurable
from stratuslab.Authn import AuthnFactory
import stratuslab.Util as Util
from stratuslab.pat.Client import PortTranslationWebClient

etree = Util.importETree()


class Monitor(Configurable):
    TEMPLATE_NIC_IP = 'template_nic_ip'
    TEMPLATE_NIC_HOSTNAME = 'template_nic_hostname'

    ipToHostname = True

    @staticmethod
    def addOptions(parser):
        parser.add_option('-n', dest='ipToHostname', action='store_false',
                          help='Don\'t do reverse DNS resolution - print IPs of instances.',
                          default=Monitor.ipToHostname)

    def __init__(self, configHolder):
        self.endpoint = None
        self.verboseLevel = 1
        self.portTranslation = False
        self.portTranslationClient = None

        super(Monitor, self).__init__(configHolder)

        self._setCloud()

        self.hostInfoDetailAttributes = (['id', 4], ['name', 16], ['im_mad', 8], ['vm_mad', 8], ['tm_mad', 8])
        self.hostInfoListAttributes = (['id', 4], ['name', 16])

        self.vmInfoDetailAttributes = [['id', 4], ['state_summary', 10], ['template_vcpu', 5], ['memory', 10],
                                       ['cpu', 5], [Monitor.TEMPLATE_NIC_HOSTNAME, 24], ['name', 16]]
        self.vmInfoListAttributes = [['id', 4], ['state_summary', 10], ['template_vcpu', 5], ['memory', 10], ['cpu', 5],
                                     [Monitor.TEMPLATE_NIC_HOSTNAME, 24], ['name', 16]]

        self.labelDecorator = {'state_summary': 'state', Monitor.TEMPLATE_NIC_HOSTNAME: 'host/ip',
                               Monitor.TEMPLATE_NIC_IP: 'ip', 'template_vcpu': 'vcpu', 'cpu': 'cpu%'}

        if Util.isTrueConfVal(self.portTranslation):
            self.portTranslationClient = PortTranslationWebClient(configHolder)
            self.vmInfoDetailAttributes.insert(-1, ['template_pat', 16])
            self.vmInfoListAttributes.insert(-1, ['template_pat', 16])
            self.labelDecorator['template_pat'] = 'pat(VM:GW)'

    def _setCloud(self):
        credentials = AuthnFactory.getCredentials(self)
        self.cloud = CloudConnectorFactory.getCloud(credentials)

        if self.endpoint:
            self.cloud.setEndpoint(self.endpoint)
        elif 'frontendIp' in self.__dict__ and 'proxyPort' in self.__dict__:
            self.cloud.setEndpointFromParts(self.frontendIp, self.proxyPort)
        else:
            self.cloud.setEndpoint(self.endpoint)

    def nodeDetail(self, nodeIds):
        infoList = []
        for nodeId in nodeIds:
            infoList.append(self._nodeDetail(nodeId))
        return infoList

    def _nodeDetail(self, nodeId):
        res = self.cloud.getHostInfo(int(nodeId))
        host = etree.fromstring(res)
        info = CloudInfo()
        info.populate(host)
        return info

    def vmDetail(self, vmIds):
        infoList = []
        for vmId in vmIds:
            infoList.append(self._vmDetail(vmId))
        return infoList

    def _vmDetail(self, vmId):
        res = self.cloud.getVmInfo(int(vmId))
        vm = etree.fromstring(res)
        if self.portTranslationClient:
            self.portTranslationClient.addPortTranslationToSingleVmInfo(vm)
        info = CloudInfo()
        info.populate(vm)
        self._addHostnameElement(info)
        return info

    def _addHostnameElement(self, info):
        ip = info.attribs[Monitor.TEMPLATE_NIC_IP]
        info.set(Monitor.TEMPLATE_NIC_HOSTNAME, self._ipv4ToHostname(ip))

    def _addHostnameToVmsInfo(self, vmsInfo):
        for vmInfo in vmsInfo.findall('VM'):
            self._addHostnameToSingleVmInfo(vmInfo)

    def _addHostnameToSingleVmInfo(self, vmInfo):
        for nic in vmInfo.findall('TEMPLATE/NIC'):
            ipv4 = nic.find('IP').text
            hostname = self._ipv4ToHostname(ipv4)

            host_element = etree.Element("HOSTNAME")
            host_element.text = hostname

            nic.append(host_element)

    def _ipv4ToHostname(self, ipv4):
        if self.ipToHostname:
            try:
                return socket.gethostbyaddr(ipv4)[0]
            except:
                pass
        return ipv4

    def listNodes(self):
        nodes = self.cloud.listHosts()
        correct_nodes = []
        for node in self._iterate(etree.fromstring(nodes)):
            # FIXME: remove this later.
            try:
                node.attribs['template_usedcpu'] = round(float(node.attribs['template_usedcpu']), 2)
            except:
                pass
            correct_nodes.append(node)
        return correct_nodes

    def listVms(self, showVmsFromAllUsers=False):
        res = self.cloud.listVms(showVmsFromAllUsers)
        vms = etree.fromstring(res)

        self._addHostnameToVmsInfo(vms)

        if self.portTranslationClient:
            self.portTranslationClient.addPortTranslationToVmsInfo(vms)

        correct_vms = []
        for vm in self._iterate(vms):
            self._addHostnameElement(vm)
            if vm.attribs['uname'].startswith('CN'):
                vm.attribs['uname'] = vm.attribs['uname'].replace("CN%3D", "") \
                    .replace("%2COU%3D", " OrgUnit:").replace("%2CO%3D", " Org:").replace("%2CC%3D", " Country:") \
                    .replace("+", " ")
            correct_vms.append(vm)
        return correct_vms

    def _iterate(self, items):
        infoList = []
        for item in items:
            info = CloudInfo()
            info.populate(item)
            infoList.append(info)
        return infoList

    def formatVmList(self, items):
        self._adjustVmAttributeFields(items)
        result = self._formatVmInfoHeader(self.vmInfoListAttributes)
        for item in items:
            result += self._formatVmInfo(item, self.vmInfoListAttributes)
        return result

    def formatVmDetails(self, items):
        self._adjustVmAttributeFields(items)
        result = self._formatVmInfoHeader(self.vmInfoDetailAttributes)
        for item in items:
            result += self._formatVmInfo(item, self.vmInfoDetailAttributes)
        return result

    def formatVmAllAttributes(self, vmInfoList):
        result = ''
        for vmInfo in vmInfoList:
            vmAttributes = vmInfo.getAttributes()
            result = '%s\n' % ('-' * 25)
            result += '%s (all attributes)\n' % vmAttributes['id']
            for k in sorted(vmAttributes.keys()):
                result += '  %s = %s\n' % (k, vmAttributes[k])
        return result

    def _formatVmInfoHeader(self, headerAttributes):
        result = Util.EMPHASIS_START

        try:
            for attrib in headerAttributes[:-1]:
                label = self._decorateLabel(attrib[0])
                result += label.ljust(int(attrib[1]))
                # adjust last element to its own length

            attrib = headerAttributes[-1]
            label = self._decorateLabel(attrib[0])
            result += label.ljust(len(attrib[0]))
        finally:
            result += Util.EMPHASIS_STOP
            result += "\n"

        return result

    def _decorateLabel(self, label):
        return self.labelDecorator.get(label, label)

    def _formatVmInfo(self, info, headerAttributes):
        result = ''
        for attrib in headerAttributes[:-1]:
            result += getattr(info, attrib[0], '').ljust(int(attrib[1]))
        result += self._formatVmName(headerAttributes, info)

        result += self._formatErrorIfRequired(info, headerAttributes)

        return result

    def _formatVmName(self, headerAttributes, info):
        # adjust last element to its own length
        attrib = headerAttributes[-1]
        return getattr(info, attrib[0]).ljust(len(attrib[0])) + "\n"

    def _formatErrorIfRequired(self, info, headerAttributes):
        result = ''
        if self.verboseLevel >= Util.VERBOSE_LEVEL_NORMAL:
            if getattr(info, 'template_error_message', None):
                result = ' ' * headerAttributes[0][1] + info.template_error_message.strip() + '\n'

        return result

    def _adjustVmAttributeFields(self, _list):
        attrList = ('vmInfoDetailAttributes', 'vmInfoListAttributes')
        self._adjustAttributeFields(_list, attrList)

    def _adjustHostAttributeFields(self, _list):
        attrList = ('hostInfoDetailAttributes', 'hostInfoDetailAttributes')
        self._adjustAttributeFields(_list, attrList)

    def _adjustAttributeFields(self, _list, attrList):
        if _list:
            for attr in attrList:
                for i, attrVal in enumerate(getattr(self, attr)):
                    lenMax = max(map(lambda x: len(getattr(x, attrVal[0], '') or ''), _list))
                    if lenMax >= getattr(self, attr)[i][1]:
                        getattr(self, attr)[i][1] = lenMax + 1

    def getVmConnectionInfo(self, vmId):
        host, port = None, None

        vmPort = str(self.cloud.getVmSshPort(int(vmId)))
        vmInfo = self._vmDetail(vmId)
        if self.portTranslationClient and self.portTranslationClient.hasPortTranslated(vmInfo):
            port = self.portTranslationClient.findGatewayPort(vmInfo, vmPort)
            host = self.patGatewayHost
        else:
            port = vmPort
            host = getattr(vmInfo, Monitor.TEMPLATE_NIC_IP, None)
        return host, port


class MultisiteMonitor(object):
    def __init__(self, configHolder):
        self.endpointTimeout = '5'
        self.configHolder = configHolder
        configHolder.assign(self)
        self.endpoints = [eup.strip().split() for eup in self.endpoints.split('\n')]

    def formatVmList(self):
        configHolder = self.configHolder.copy()

        result = ''
        for endpoint, username, password in self.endpoints:
            result += Util.EMPHASIS_START
            result += '::: %s : %s :::' % (endpoint, username)
            result += Util.EMPHASIS_STOP

            configHolder.set('endpoint', endpoint)
            configHolder.set('username', username)
            configHolder.set('password', password)
            monitor = Monitor(configHolder)
            self._set_alarm()
            try:
                vmList = monitor.listVms()
                result += monitor.formatVmList(vmList)
            except Exception, ex:
                Util.printError(str(ex))
            except KeyboardInterrupt:
                pass
            self._unset_alarm()

        return result

    def _set_alarm(self):
        def _sig_handler(_ignore1, _ignore2):
            raise Exception("WARNING: Could not get results in %s s" %
                            self.endpointTimeout)

        signal.signal(signal.SIGALRM, _sig_handler)
        signal.alarm(int(self.endpointTimeout))

    def _unset_alarm(self):
        signal.alarm(0)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from stratuslab.system.centos import CentOS
from stratuslab.system.PackageInfo import PackageInfo

class Fedora(CentOS):
    
    os = 'fedora'
    
    def __init__(self):
        super(Fedora, self).__init__()
        self.frontendDeps = [
                'openssh', 'ruby', 'zlib-devel', 'curl'
        ]
        
        self.packages['dhcp'] = PackageInfo('dhcp',
                                            configFile='/etc/dhcp/dhcpd.conf',
                                            initdScriptName='dhcpd')
        

system = Fedora()



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2010-11, GRNET S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import socket
import time

from stratuslab.Util import printStep, printAction, printError, printInfo, printDebug


class Host(object):
    def __init__(self):
        self.public_ip = None
        self.public_dns = None
        self.cores = None
        self.ram = None
        self.swap = None


class SSHUtil(object):
    _options = " -q -o StrictHostKeyChecking=no "
    _private_key = None
    _username = None

    def __init__(self, private_key, username):
        self._private_key = private_key
        self._username = username

    def copy_file_to_hosts(self, hostlist, srcfile, remotepath):
        cmd = "scp" + self._options + " -i " + self._private_key

        for host in hostlist:
            printDebug('Copying to host %s' % host.public_ip)
            error = os.system(cmd + " " + srcfile + " " + self._username + "@" + host.public_ip + ":" + remotepath)
            if error > 0:
                return error

        return 0

    def run_remote_command(self, hostlist, command):
        for host in hostlist:
            cmd = "ssh -y" + self._options + " -i " + self._private_key + " " + \
                  self._username + "@" + host.public_ip + " " + command
            printDebug('Command: %s' % cmd)
            printDebug('Executing in host %s' % host.public_ip)
            error = os.system(cmd)
            if error > 0:
                return error

        return 0

    def waitForConnectivity(self, host, timeout):
        start_time = time.time()

        additional_options = "-o ConnectTimeout=10"

        cmd = "ssh" + self._options + " " + additional_options + " -i " + self._private_key + " " + \
              self._username + "@" + host.public_ip + " true"

        error = 1

        while (time.time() - start_time) < timeout and error > 0:
            error = os.system(cmd)

        if error > 0:
            return False
        else:
            return True


class Cluster(object):
    #hosts = []
    #_is_heterogeneous = False

    def __init__(self, configHolder, runner, master_vmid):
        configHolder.assign(self)
        self._runner = runner
        self.hosts = []
        self._is_heterogeneous = False
        if master_vmid:
            self._master_vmid = master_vmid
            self._is_heterogeneous = True

    def create_machine_file(self, hostlist, filename, isForMPI=False):
        mf = open(filename, "w")

        for host in hostlist:
            if isForMPI:
                additionalInfo = " slots=" + str(host.cores)
            else:
                additionalInfo = ""

            mf.write(host.public_ip + additionalInfo + "\n")
        mf.close()

    def doAddPackages(self, ssh):
        printStep('Installing additional software packages')
        packages = self.add_packages.replace(",", " ")

        printStep('Trying to configure new apps with yum...')
        if ssh.run_remote_command(self.hosts, "yum -q -y install " + packages):
            printStep('Trying to configure new apps with apt-get...')
            ssh.run_remote_command(self.hosts, "apt-get -q -y install " + packages)

    def doPrepareMPImachineFile(self, ssh, worker_nodes):
        printStep('Preparing MPI machine file')
        if self.include_master:
            target = self.hosts
        else:
            target = worker_nodes

        self.create_machine_file(target, "/tmp/machinefile", isForMPI=True)
        ssh.copy_file_to_hosts(self.hosts, "/tmp/machinefile", "/tmp")
        os.unlink("/tmp/machinefile")

    def doPrepareNodeList(self, ssh, worker_nodes):
        if self.include_master:
            target = self.hosts
        else:
            target = worker_nodes

        self.create_machine_file(target, "/tmp/cluster_nodelist")
        ssh.copy_file_to_hosts(self.hosts, "/tmp/cluster_nodelist", "/tmp")
        os.unlink("/tmp/cluster_nodelist")

    def doPrepareNFSSharedFolder(self, ssh, master_node, worker_nodes):
        printStep('Preparing NFS shared folder')
        master_only = []
        master_only.append(master_node)
        ssh.run_remote_command(self.hosts, "mkdir -p " + self.shared_folder)
        ssh.run_remote_command(master_only,
                               "'echo " + self.shared_folder + " \"*(rw,no_root_squash)\" >> /etc/exports'")

        printStep('\tTrying RedHat configuration...')
        if ssh.run_remote_command(master_only, "service nfs restart &> /dev/null"):
            printStep('\tTrying debian configuration...')
            ssh.run_remote_command(master_only, "service nfs-kernel-server restart &> /dev/null")

        ssh.run_remote_command(worker_nodes,
                               "mount " + master_node.public_ip + ":" + self.shared_folder + " " + self.shared_folder)

    def doSetupSSHHostBasedCluster(self, ssh):
        printStep('Configuring passwordless host-based ssh authentication')
        ssh.run_remote_command(self.hosts,
                               "'echo \"IgnoreRhosts no\" >> /etc/ssh/sshd_config && service sshd restart &> /dev/null && " +
                               "echo \"HostbasedAuthentication yes\n" +
                               "StrictHostKeyChecking no\n" +
                               "EnableSSHKeysign yes\" >> /etc/ssh/ssh_config'")

        for host in self.hosts:
            ssh.run_remote_command(self.hosts,
                                   "'ssh-keyscan -t rsa " + host.public_dns + " 2>/dev/null >> /etc/ssh/ssh_known_hosts && " \
                                                                              "echo " + host.public_dns + " root >> /root/.shosts'")

    def doCreateClusterUser(self, ssh, master_node):
        printStep('Creating additional user')
        master_only = []
        master_only.append(master_node)
        ssh.run_remote_command(self.hosts, "useradd -m " + self.cluster_user)
        ssh.run_remote_command(master_only,
                               ' "su - ' + self.cluster_user + " -c 'ssh-keygen -q -t rsa -N " + '\\"\\"' " -f ~/.ssh/id_rsa' " + '"')
        ssh.run_remote_command(master_only,
                               ' "su - ' + self.cluster_user + " -c 'cp ~/.ssh/id_rsa.pub ~/.ssh/authorized_keys' " + '"')

        #if self.shared_folder !="home":
        #    for host in self.hosts:
        #        ssh.run_remote_command(master_only, "scp -r /home/"+ self.cluster_user+"/.ssh " + host.public_ip + ":/home/" + self.cluster_user)

        if self.ssh_hostbased:
            for host in self.hosts:
                ssh.run_remote_command(self.hosts,
                                       ' "su - ' + self.cluster_user + " -c 'echo " + host.public_dns + " " + self.cluster_user + " >> ~/.shosts'" + '"')


    def doUpdateEnvironmentVariables(self, ssh, master_node, worker_nodes):
        printStep('Updating environment variables')
        active_nodes = []
        if self.include_master:
            active_nodes = self.hosts
        else:
            active_nodes = worker_nodes

        # Find the total available number of cores
        total_cores = 0
        for node in active_nodes:
            total_cores += int(node.cores)

        counter = 0
        for node in self.hosts:
            target = []
            target.append(node)
            ssh.run_remote_command(target, "'echo export STRATUSLAB_NC=" + str(
                counter) + " > /etc/profile.d/stratuslab_cluster.sh && " \
                           "echo export STRATUSLAB_CMASTER=" + master_node.public_dns + " >> /etc/profile.d/stratuslab_cluster.sh && " \
                                                                                        "echo export STRATUSLAB_CSIZE=" + str(
                len(active_nodes)) + " >> /etc/profile.d/stratuslab_cluster.sh && " \
                                     "echo export STRATUSLAB_CMAX_CORES=" + str(
                total_cores) + " >> /etc/profile.d/stratuslab_cluster.sh'")
            counter += 1


    def doUpdateHostsFile(self, ssh, master_node, worker_nodes):
        printStep('Updating hosts file')
        ssh.run_remote_command(self.hosts, "'echo  >> /etc/hosts && " \
                                           " echo \"# Cluster nodes\" >> /etc/hosts && " \
                                           " echo " + master_node.public_ip + " " + master_node.public_dns + " " + "master >> /etc/hosts'")

        counter = 0
        for host in worker_nodes:
            ssh.run_remote_command(self.hosts, " 'echo " + host.public_ip + " " + host.public_dns + " worker-" + str(
                counter) + " >> /etc/hosts'")
            counter += 1

    def doStartClusterServices(self, ssh, master_node):
        printStep("Applying user defined cluster services")
        master_only = []
        master_only.append(master_node)
        ssh.run_remote_command(master_only, "'if [ -e /etc/rc.cluster-services ]; then /etc/rc.cluster-services; fi'")

    def deploy(self):
        ssh = SSHUtil(self._runner.userPrivateKeyFile, self.cluster_admin)

        # Wait until all the images are up and running
        vmNetworkDetails = []
        vmStartTimeout = 600

        # wait until the each machine is up or timeout after 15 minutes
        printStep("Waiting for all cluster VMs to be instantiated...")
        if self._is_heterogeneous:
            printStep("Waiting for master")
            self._runner.waitUntilVmRunningOrTimeout(self._master_vmid, vmStartTimeout)
            vmNetworkDetails.append(self._runner.getNetworkDetail(self._master_vmid))

        for vmId in self._runner.vmIds:
            printDebug('Waiting for instance to start running %s' % str(vmId))
            self._runner.waitUntilVmRunningOrTimeout(vmId, vmStartTimeout)
            vmNetworkDetails.append(self._runner.getNetworkDetail(vmId))

        vm_cpu, vm_ram, vm_swap = self._runner.getInstanceResourceValues()

        for vmNetwork in vmNetworkDetails:
            if vmNetwork[0] == 'public':
                host = Host()
                host.public_ip = vmNetwork[1]

                try:
                    host.public_dns = socket.gethostbyaddr(host.public_ip)[0]
                except:
                    host.public_dns = host.public_ip

                host.cores = vm_cpu
                host.ram = vm_ram
                host.swap = vm_swap
                self.hosts.append(host)

        printStep("Waiting for all instances to become accessible...")

        failedHosts = []

        for host in self.hosts:
            hostReady = False
            hostFailed = False

            while not hostReady and not hostFailed:
                if not ssh.waitForConnectivity(host, vmStartTimeout):
                    printError('Timed out while connecting to %s.  Removing from target config. list.' % host.public_ip)
                    failedHosts.append(host)
                    hostFailed = True
                else:
                    hostReady = True

        if len(failedHosts) > 0:
            if self.tolerate_failures:
                for host in failedHosts:
                    self.hosts.remove(host)
            else:
                printError('Error instantiating some or all of the nodes. Bailing out...')
                if self.clean_after_failure:
                    self._runner.killInstances(self._runner.vmIds)
                return 128

        master_node = self.hosts[0]

        worker_nodes = list(self.hosts)

        worker_nodes.remove(master_node)

        printInfo('\tMaster is %s' % master_node.public_dns)

        for node in worker_nodes:
            printInfo('\tWorker: %s' % node.public_dns)

        # Configure the hosts
        printAction('Configuring nodes')

        # Try to install the missing packages
        if self.add_packages:
            self.doAddPackages(ssh)

        # For MPI clusters prepare the machinefile for mpirun
        if self.mpi_machine_file:
            self.doPrepareMPImachineFile(ssh, worker_nodes)

        if self.cluster_user:
            # Create a new user and prepare the environments for password-less ssh
            self.doCreateClusterUser(ssh, master_node)

        # Initialize the shared storage in NFS
        if self.shared_folder:
            self.doPrepareNFSSharedFolder(ssh, master_node, worker_nodes)

        if self.ssh_hostbased:
            self.doSetupSSHHostBasedCluster(ssh)

        # Update /etc/profile with StratusLab specific environment variables
        self.doUpdateEnvironmentVariables(ssh, master_node, worker_nodes)

        # Store the list of cluster nodes in a file under /tmp
        self.doPrepareNodeList(ssh, worker_nodes)

        # Update the /etc/hosts file for all hosts
        self.doUpdateHostsFile(ssh, master_node, worker_nodes)

        # Start any services defined in rc.cluster-services
        self.doStartClusterServices(ssh, master_node)

        return 0



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, Centre National de la Recherche Scientifique (CNRS)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest
import httplib2
import urllib
import random
import string

import ldap
import ldap.modlist as modlist

class RegistrationTest(unittest.TestCase):

    baseUrl = 'https://localhost:8444'
    ldapUrl = 'ldaps://localhost:636/'

    managerDn = 'cn=admin,o=cloud'
    managerPassword = 'secret'

    username = 'jsmith'
    userDn = 'uid=jsmith,ou=users,o=cloud'
    userPassword = 'dummy_password'
    userPasswordNew = 'dummy_password1'
    
    baseDns = { 'users':'ou=users,o=cloud', 
                'groups':'ou=groups,o=cloud',
                'actions':'ou=actions,o=cloud' } 

    headers = {'cache-control':'no-cache'}

    def getLdapResultSet(self, ldapObj, ldapResultId):
        resultSet = []
        
        while 1:
            resultType, resultData = ldapObj.result(ldapResultId, 0)
            if (resultData != []):
                if resultType == ldap.RES_SEARCH_ENTRY:
                    resultSet.append(resultData)
            else:
                break
        
        return resultSet


    def pageExistsWithNonZeroLength(self, url):
        h = httplib2.Http(".cache", disable_ssl_certificate_validation=True)
        resp, content = h.request(url);

        self.assertEquals(200, resp.status, url + ': invalid status ' + str(resp.status))

        length = len(content)
        self.assertTrue(length > 0, url + ': invalid length ' + str(length))


    def registerUseCase(self):
        h = httplib2.Http(disable_ssl_certificate_validation=True)
        h.follow_redirects = False

        data = {
            'uid':self.username, 
            'mail':'smith@example.org',
            'givenName':'john',
            'sn':'smith',
            'seeAlso':'cn=John Smith,o=StratusLab',
            'newUserPassword':self.userPassword,
            'newUserPasswordCheck':self.userPassword,
            'message':'I love clouds',
            'agreement':'true'
            }

        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        body = urllib.urlencode(data)
        
        url = self.baseUrl + '/users/'
        resp, content = h.request(url, "POST", body=body, headers=headers)

        # A correct registration should return a redirect. 
        self.assertEquals(303, resp.status, url + ': invalid status ' + str(resp.status))

        # Ensure user is found in LDAP
        self.findUserInLdap()


    def findUserInLdap(self):

        ldapObj = ldap.initialize(self.ldapUrl)
        ldapObj.simple_bind_s(self.managerDn, self.managerPassword)
        
        searchScope = ldap.SCOPE_BASE
        searchFilter = 'objectClass=inetOrgPerson'
        retrieveAttributes = None

        ldapResultId = ldapObj.search(self.userDn, searchScope, 
                                      searchFilter, retrieveAttributes)
        
        resultSet = self.getLdapResultSet(ldapObj, ldapResultId)
        
        self.assertEqual(1, len(resultSet), 'user not found in LDAP: ' + self.userDn)


    def checkUserLdapPermissions(self):

        ldapObj = ldap.initialize(self.ldapUrl)
        ldapObj.simple_bind_s(self.userDn, self.userPassword)
        
        searchScope = ldap.SCOPE_BASE
        searchFilter = 'objectClass=*'
        retrieveAttributes = None

        for key in self.baseDns.keys():
            baseDn = self.baseDns[key]
            ldapResultId = ldapObj.search(baseDn, searchScope, 
                                          searchFilter, retrieveAttributes)

            resultSet = self.getLdapResultSet(ldapObj, ldapResultId)
                    
            self.assertEqual(1, len(resultSet), 'user cannot read base object: ' + baseDn)


    def profileExistsWithNonZeroLength(self):
        h = httplib2.Http(".cache", disable_ssl_certificate_validation=True)
        h.add_credentials(self.username, self.userPassword)
        resp, content = h.request(self.baseUrl + '/profile/')

        self.assertEquals(200, resp.status, url + ': invalid status ' + str(resp.status))

        length = len(content)
        self.assertTrue(length > 0, url + ': invalid length ' + str(length))


    def startPasswordReset(self):
        h = httplib2.Http(disable_ssl_certificate_validation=True)
        h.follow_redirects = False

        data = {
            'uid':self.username
            }

        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        body = urllib.urlencode(data)
        
        url = self.baseUrl + '/reset/'
        resp, content = h.request(url, "POST", body=body, headers=headers)

        # A correct registration should return a redirect. 
        self.assertEquals(303, resp.status, url + ': invalid status ' + str(resp.status))


    def getActionsFromLdap(self):

        ldapObj = ldap.initialize(self.ldapUrl)
        ldapObj.simple_bind_s(self.managerDn, self.managerPassword)
        
        searchScope = ldap.SCOPE_SUBTREE
        searchFilter = 'objectClass=javaSerializedObject'
        retrieveAttributes = ['cn']

        baseDn = self.baseDns['actions']
        ldapResultId = ldapObj.search(baseDn, searchScope, 
                                      searchFilter, retrieveAttributes)
        
        resultSet = self.getLdapResultSet(ldapObj, ldapResultId)

        actionCnSet = set()
        for item in resultSet:
            actionCnSet.add(item[0][1]['cn'][0])

        return actionCnSet


    def confirmPasswordReset(self, actionUuid):
        h = httplib2.Http(disable_ssl_certificate_validation=True)

        url = self.baseUrl + '/action/' + actionUuid
        resp, content = h.request(url)

        self.assertEquals(200, resp.status, url + ': invalid status ' + str(resp.status))


    def resetPasswordUseCase(self):

            # Try to reset the user password.
            previousActions = self.getActionsFromLdap()
            self.startPasswordReset()
            currentActions = self.getActionsFromLdap()
            newActions = currentActions.difference(previousActions)
            self.assertEquals(1, len(newActions))
            actionUuid = newActions.pop()
            self.confirmPasswordReset(actionUuid)

            # Accessing service with old password should fail.
            try:
                self.checkUserLdapPermissions()
            except ldap.INVALID_CREDENTIALS:
                # Ok, tried to connect with wrong password.
                consumed = True


    def deleteUserFromLdap(self):
        ldapObj = ldap.initialize(self.ldapUrl)
        ldapObj.simple_bind_s(self.managerDn, self.managerPassword)
        
        ldapObj.delete_s(self.userDn)


    def testRegistrationServicePagesExist(self):
        urls = [ '/', '/register/', '/policies/', '/reset/' ]
        for url in urls:
            self.pageExistsWithNonZeroLength(self.baseUrl + url);
        

    def testLdapBaseObjectsExist(self):

        ldapObj = ldap.initialize(self.ldapUrl)
        ldapObj.simple_bind_s(self.managerDn, self.managerPassword)
        
        searchScope = ldap.SCOPE_BASE
        searchFilter = 'objectClass=*'
        retrieveAttributes = None

        for key in self.baseDns.keys():
            baseDn = self.baseDns[key]
            ldapResultId = ldapObj.search(baseDn, searchScope, 
                                          searchFilter, retrieveAttributes)

            resultSet = self.getLdapResultSet(ldapObj, ldapResultId)
                    
            self.assertEqual(1, len(resultSet), 'problem with base object: ' + baseDn)


    def testRegisterUser(self):
        try:

            self.registerUseCase()

            self.checkUserLdapPermissions()

            self.resetPasswordUseCase()

        finally:                
            self.deleteUserFromLdap()

if __name__ == "__main__":
    unittest.main()

